# SPDX-License-Identifier: Apache-2.0

"""
Tests for the shared factory implementation
"""
# NOTE: We use lots of "bad" names in tests to indicate that they're not real!
# pylint: disable=disallowed-name

# Standard
from unittest import mock
import abc

# Third Party
from jsonschema.exceptions import ValidationError
import aconfig
import pytest

# Local
from granite_io import factory

## Helpers #####################################################################


class BaseDummy(factory.FactoryConstructible):
    """Base class for dummy classes"""

    def __init__(self, config: aconfig.Config):
        self._config = config

    @abc.abstractmethod
    def doit(self, it: str) -> str:
        """Do something"""


## Tests #######################################################################


def test_factory_decorate_and_construct_class():
    """Test using the decorator on classes works to register and construct"""
    dummy_factory = factory.Factory("dummy")
    assert dummy_factory.name == "dummy"
    assert not dummy_factory.registered_types()

    @dummy_factory.decorator("dummy-one", "Dummy One")
    class DummyOne(BaseDummy):
        def doit(self, it: str) -> str:
            return f"Doing ONE it: {it}"

    @dummy_factory.decorator("dummy-two", "d2")
    class DummyTwo(BaseDummy):  # pylint: disable=unused-variable
        def doit(self, it: str) -> str:
            return f"Doing TWO its: {it} {it}"

    assert len(dummy_factory.registered_types()) == 4

    d1_a = dummy_factory.construct("dummy-one")
    assert isinstance(d1_a, DummyOne)
    d1_b = dummy_factory.construct("Dummy One")
    assert isinstance(d1_b, DummyOne)


def test_construct_with_config():
    """Test that constructing an object with a config works as expected."""
    dummy_factory = factory.Factory("dummy")

    @dummy_factory.decorator("dummy-one", "Dummy One")
    class DummyOne(BaseDummy):
        def doit(self, it: str) -> str:
            return f"Doing ONE it: {it}"

    d1 = dummy_factory.construct({"type": "dummy-one"})
    assert isinstance(d1, DummyOne)


def test_decorate_function():
    """Test that a function can be decorated."""
    dummy_factory = factory.Factory("dummy")

    @dummy_factory.decorator(
        "d1",
        config_schema={
            "properties": {"foo": {"type": "string"}, "bar": {"type": "string"}}
        },
        config_defaults={"bar": "BAR"},
    )
    def d(it: str, foo: str, bar: str) -> str:
        return f"Foo: {foo}, Bar: {bar}, it: {it}"

    d1 = dummy_factory.construct("d1", {"foo": "FOO"})
    assert d1("it") == "Foo: FOO, Bar: BAR, it: it"


def test_decorate_function_config_mismatch():
    """Make sure that the config schema must match the function signature"""
    dummy_factory = factory.Factory("dummy")

    def d1(it: str, foo: str, bar: str) -> str:
        return f"Foo: {foo}, Bar: {bar}, it: {it}"

    with pytest.raises(ValueError):
        dummy_factory.decorator(
            "d1",
            config_schema={"properties": {"buz": {"type": "string"}}},
            config_defaults={"bar": "BAR"},
        )(d1)


def test_decorate_function_bad_type():
    """Make sure that decorating something that isn't a class or function raises
    an error
    """
    dummy_factory = factory.Factory("dummy")
    x = 1
    with pytest.raises(TypeError):
        dummy_factory.decorator("d1")(x)


def test_nested_config():
    """Make sure that nested config defaults and types are handled correctly"""
    dummy_factory = factory.Factory("dummy")

    @dummy_factory.decorator(
        "d",
        config_defaults={"foo": {"bar": 1}},
        config_schema={
            "type": "object",
            "properties": {
                "foo": {
                    "type": "object",
                    "properties": {
                        "bar": {
                            "type": "number",
                        },
                    },
                },
            },
        },
    )
    def d(foo: dict, bar_exp: int) -> str:
        assert "bar" in foo
        assert foo["bar"] == bar_exp

    dummy_factory.construct("d")(bar_exp=1)
    dummy_factory.construct("d", {"foo": {"bar": 2}})(bar_exp=2)


def test_config_schema_validation():
    """Make sure that the config schema must match the function signature"""
    dummy_factory = factory.Factory("dummy")

    @dummy_factory.decorator("d")
    def d(foo: int):
        return foo

    with pytest.raises(ValidationError):
        dummy_factory.construct("d", {"foo": "bar"})


def test_schema_inference_with_defaults():
    """Make sure the config schema and defaults can be correctly inferred"""
    dummy_factory = factory.Factory("dummy")

    @dummy_factory.decorator("d")
    def d(foo: int = 1, bar: str = "asdf"):
        return f"{bar}[{foo}]"

    assert dummy_factory.construct("d")() == "asdf[1]"
    assert dummy_factory.construct("d", {"foo": 2})() == "asdf[2]"


def test_schema_inference_no_hint():
    """Make sure schema inference raises if the function doesn't have type hints"""
    dummy_factory = factory.Factory("dummy")

    def d(foo):
        return foo

    with pytest.raises(ValueError):
        dummy_factory.decorator("d")(d)


def test_merge_configs_corners():
    """Make sure merge config corner cases are handled correctly"""
    base = {"a": 1, "b": 2}
    overrides = {"a": 11, "c": 3}
    assert factory._merge_configs(None, overrides) is overrides
    assert factory._merge_configs(base, None) is base
    assert factory._merge_configs(base, overrides) == {"a": 11, "b": 2, "c": 3}


def test_importable_factory():
    """Test that registration can happen with a dynamic import using
    ImportableFactory
    """
    dummy_ifactory = factory.ImportableFactory("dummy")

    class DummyOne(BaseDummy):
        name = "d1"

        def doit(self, it: str) -> str:
            return f"ONE[{it}]"

    mock_module = aconfig.Config({"DummyOne": DummyOne}, override_env_vars=False)

    with mock.patch("importlib.import_module", return_value=mock_module) as import_mock:
        # Without the import keyword, it should fail to construct
        with pytest.raises(ValueError):
            dummy_ifactory.construct({"type": "d1"})
        assert import_mock.call_count == 0

        # With the import keyword, it should succeed
        dummy_one = dummy_ifactory.construct(
            {"type": "d1", dummy_ifactory.IMPORT_CLASS_KEY: "foo.bar.DummyOne"}
        )
        assert isinstance(dummy_one, DummyOne)
        assert import_mock.call_count == 1


def test_bad_construct_arg_type():
    """Test that calling construct with a bad type raises"""
    dummy_factory = factory.Factory("dummy")
    with pytest.raises(TypeError):
        dummy_factory.construct(1)



# SPDX-License-Identifier: Apache-2.0

"""
Simple placeholder for unit testing contrib
"""

# Third Party
from src.hello_lib import hello_world


def test_hello():
    """Simple example test"""
    ret = hello_world()
    assert ret == "Hello today!"  # pylint: disable=comparison-of-constants

    # pylint: disable-next=comparison-of-constants, comparison-with-itself
    assert "Hello today!" == "Hello today!"



# SPDX-License-Identifier: Apache-2.0

# Standard
import os
import re

# Third Party
from litellm import UnsupportedParamsError
import pytest

# Local
from granite_io import make_backend
from granite_io.backend.openai import OpenAIBackend
from granite_io.types import GenerateInputs, GenerateResults


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_simple(backend_x):
    ret = await backend_x(GenerateInputs(prompt="hello"))
    assert isinstance(ret, GenerateResults)
    assert len(ret.results) == 1


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_num_return_sequences_1(backend_x):
    ret = await backend_x(GenerateInputs(prompt="hello", n=1))
    assert isinstance(ret, GenerateResults)
    assert len(ret.results) == 1


@pytest.mark.asyncio
@pytest.mark.vcr
async def test_num_return_sequences_3(backend_x):
    try:
        ret = await backend_x(GenerateInputs(prompt="what is up?", n=3))
    except UnsupportedParamsError:
        # Specific exception from LiteLLMBackend
        # xfail because LiteLLM is telling us that ollama does not support
        # n > 1, but we can use LiteLLM with watsonx FTW
        pytest.xfail("LiteLLMBackend support for n > 1 varies by provider")

    num_returned = len(ret.results)

    if num_returned == 1 and isinstance(backend_x, OpenAIBackend):
        # ollama with OpenAI will just return 1, other OpenAI backends can return 3
        pytest.xfail("OpenAIBackend support for n > 1 varies by provider")

    assert isinstance(ret, GenerateResults)
    assert num_returned == 3


@pytest.mark.skipif(not os.environ.get("OPENAI_API_KEY"), reason="Needs OpenAI API key")
async def test_extra_headers():
    """Test extra_headers using manual test case against a server to verify.

    Use environment variables suitable for an OpenAI backend:
    OPENAI_BASE_URI - base URL to the provider
    MODEL_NAME - model name override for the provider
    OPENAI_API_KEY - api key used for the provider
    Note:  For this specific test, we have a provider that needs
           the API key to also be in an additional header.
           It also passes the num_returned test (Ollama does not).
    """
    be = make_backend(
        "openai",
        {
            "model_name": "ibm-granite/granite-8b-instruct-preview-4k",
            "openai_api_key": "<your api key>",  # Use env var to set
            "openai_base_url": "<your base url>",
        },
    )

    extra_headers = {"RITS_API_KEY": os.environ.get("OPENAI_API_KEY")}
    ret = await be(
        GenerateInputs(prompt="what is up?", n=3, extra_headers=extra_headers)
    )
    num_returned = len(ret.results)
    assert isinstance(ret, GenerateResults)
    assert num_returned == 3  # Note: This would fail if you just hit Ollama


@pytest.mark.asyncio
@pytest.mark.vcr
@pytest.mark.parametrize("n", [-1, 0])
async def test_num_return_sequences_invalid(backend_x, n):
    with pytest.raises(ValueError, match=re.escape(f"Invalid value for n ({n})")):
        _ = await backend_x(inputs=GenerateInputs(prompt="hello", n=n))



# SPDX-License-Identifier: Apache-2.0

"""
Sequential Scaling I/O processor
"""

# Standard
from typing import Callable
import pathlib

# Third Party
import aconfig

# Local
from .pdl_io import PdlInputOutputProcessor
from granite_io.types import ChatCompletionResults


class SequentialScalingInputOutputProcessor(PdlInputOutputProcessor):
    """
    Input-output processor asking multiple answers until a predicate is satisfied.
    """

    def __init__(
        self,
        config: aconfig.Config | None = None,
        model: str | None = None,
        backend: str | None = None,
        max_iterations: int = 5,
        validator: Callable[[ChatCompletionResults], bool] | None = None,
    ):
        """
        :param config: Setup config for this IO processor
        :param model: Model name used by the backend.
        :param backend: Backend name.
        :param max_iterations: Maximal number of model calls.
        :param validator: predicate that the response must satisfy.
        """
        cwd = pathlib.Path(__file__).parent.resolve()
        super().__init__(
            config,
            pdl_file=cwd / "sequential_scaling.pdl",
            pdl_scope={
                "model": model,
                "backend": backend,
                "k": max_iterations,
                "validator": validator,
            },
        )



# SPDX-License-Identifier: Apache-2.0

"""
Prompt Description Language (PDL) I/O processor
"""

# Standard
from typing import Any

# Third Party
import aconfig

# Local
from granite_io.io.base import InputOutputProcessor
from granite_io.optional import import_optional
from granite_io.types import ChatCompletionInputs, ChatCompletionResults


class PdlInputOutputProcessor(InputOutputProcessor):
    """
    Base class for input-output processors that work by executing a
    Prompt Declaration Language (PDL) function processing `ChatCompletionInputs`.
    """

    with import_optional("pdl"):
        # Third Party
        # pylint: disable-next=import-error,import-outside-toplevel
        from pdl.pdl_ast import FunctionBlock, PdlLocationType

    _pdl_function: FunctionBlock | None
    """PDL function used to process the input."""
    _pdl_scope: dict[str, Any] | None
    """Initial scope in which the PDL program is executed"""
    _pdl_loc: PdlLocationType
    """Location information in the source code of the PDL function."""

    def __init__(
        self,
        config: aconfig.Config | None = None,
        pdl=None,
        pdl_file: str | None = None,
        pdl_scope: dict[str, Any] | None = None,
    ):
        """
        :param config: Setup config for this IO processor
        :param pdl: PDL function processing the `ChatCompletionInputs`
        :param pdl_file: Name of the PDL file containing the function PDL function
        :param pdl_scope: Initial environment in which the function is executed
        """
        with import_optional("pdl"):
            # Third Party
            from pdl.pdl import parse_file, parse_str
        super().__init__(config)
        if isinstance(pdl, str):
            prog, loc = parse_str(pdl, file_name=pdl_file)
            self._pdl_function = prog.root
            self._pdl_loc = loc
        elif pdl is None and pdl_file is not None:
            prog, loc = parse_file(pdl_file)
            self._pdl_function = prog.root
            self._pdl_loc = loc
        else:
            self._pdl_function = pdl
            self._pdl_loc = None
        self._pdl_scope = pdl_scope

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        return self.create_chat_completion(inputs)

    def create_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        with import_optional("pdl"):
            # Third Party
            from pdl.pdl import exec_program
            from pdl.pdl_ast import CallBlock, Program
        prog = Program(
            CallBlock(
                defs={"_pdl_function": self._pdl_function},
                call="${_pdl_function}",
                args={"inputs": inputs.model_dump()},
            )
        )
        results = exec_program(prog, scope=self._pdl_scope)
        return ChatCompletionResults.model_validate(results)



# SPDX-License-Identifier: Apache-2.0

"""
Registry of all Input, Output & IO Processors
"""

# Local
from granite_io.factory import ImportableFactory

_IO_FACTORY = ImportableFactory("io")
io_processor = _IO_FACTORY.decorator
make_io_processor = _IO_FACTORY.construct

_INPUT_FACTORY = ImportableFactory("input")
input_processor = _INPUT_FACTORY.decorator
get_input_processor = _INPUT_FACTORY.construct

_OUTPUT_FACTORY = ImportableFactory("output")
output_processor = _OUTPUT_FACTORY.decorator
get_output_processor = _OUTPUT_FACTORY.construct



# SPDX-License-Identifier: Apache-2.0


# Standard
from collections.abc import Callable
from functools import lru_cache
from typing import Union

# Third Party
from rouge_score import rouge_scorer

# Local
from granite_io.io.base import ChatCompletionResults, InputOutputProcessor
from granite_io.types import ChatCompletionInputs

scorer = rouge_scorer.RougeScorer(["rougeL"], use_stemmer=True)


def _minimum_bayesian_risk_decoding(
    answers: list[str], similarity: Callable[[str, str], float]
) -> Union[str, int, list[float]]:
    """Minimum Bayes Risk (MBR) decoding is a method for choosing the outputs of a
    machine learning system based not on the output with the highest probability, but
    the output with the lowest risk (expected error) among multiple candidates.

    In this function, it will compare each of the completion answers against each other,
    to ascertain the most similar answers.

    :param answers: List of model answers for a prompt
    :param similarity: Similarity algoritm

    :returns: Model answer most suitable
    :returns: Index from list of answers where majority answer is
    :returns: Similarity scores for all answers
    """

    similarity_scores: list[float] = []
    for _, x in enumerate(answers):
        # compute score for this answer
        score = 0
        for _, y in enumerate(answers):
            score += similarity(y, x)
        # normalization is not needed, but it gives more intepretable scores
        score = score / len(answers)
        # store
        similarity_scores.append(score)

    # select the element with highest similarity to any other
    final_index = max(range(len(answers)), key=lambda i: similarity_scores[i])
    return answers[final_index], final_index, similarity_scores


@lru_cache
def _rouge_similarity(hypotheis: str, reference: str) -> float:
    """ROUGE (Recall-Oriented Understudy for Gisting Evaluation) is used for
    evaluating the similarity between different model outputs.

    :param hypotheis: Item to compare
    :param reference: Item to compare against

    :returns: Similarity scores for the comparison
    """
    return scorer.score(hypotheis, reference)["rougeL"].fmeasure


class MBRDMajorityVotingProcessor(InputOutputProcessor):
    """
    Implementation of MBRD decoding using ROUGE score sampling for similarity
    checking.
    """

    _generator: InputOutputProcessor

    def __init__(
        self,
        generator: InputOutputProcessor,
    ):
        """
        :param generator: Sub-processor over which this processor should sample
        """
        self._generator = generator

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        # Step 1: Collect samples from the "generator" sub-processor
        base_results = await self._generator.acreate_chat_completion(inputs)

        # Step 2: Extract the content
        sample_contents = [s.next_message.content for s in base_results.results]

        # Step 3: Perform MBRD decoding and ROUGE scoring
        _, result_index, _ = _minimum_bayesian_risk_decoding(
            sample_contents, _rouge_similarity
        )

        return ChatCompletionResults(results=[base_results.results[result_index]])



# SPDX-License-Identifier: Apache-2.0

"""
The input processor module holds the abstraction and implementations for
Granite 3.3 input processing.
"""

# Local
import granite_io.io.granite_3_3.input_processors.granite_3_3_input_processor  # noqa: F401



# SPDX-License-Identifier: Apache-2.0

__doc__ = f"""
The {__package__} library provides a common programming abstraction for using
the advanced control instructions with Granite models.
"""

# Local
from granite_io.backend import Backend, backend, make_backend  # noqa: F401
from granite_io.io import (  # noqa: F401
    get_input_processor,
    get_output_processor,
    input_processor,
    io_processor,
    make_io_processor,
    output_processor,
)
from granite_io.io.base import (  # noqa: F401
    ModelDirectInputOutputProcessor,
    make_new_io_processor,
)
from granite_io.types import (  # noqa: F401
    ChatCompletionInputs,
    ChatCompletionResults,
    GenerateResult,
    GenerateResults,
    UserMessage,
)



# SPDX-License-Identifier: Apache-2.0


# Local
from .query_rewrite import QueryRewriteIOProcessor

# Expose public symbols at `granite_io.io.hallucinations` to save users from typing
__all__ = [
    "QueryRewriteIOProcessor",
]



# SPDX-License-Identifier: Apache-2.0


# Local
from . import util
from .retrieval import (
    InMemoryRetriever,
    RetrievalRequestProcessor,
    Retriever,
    compute_embeddings,
    write_embeddings,
)

# Expose public symbols at `granite_io.io.retrieval` to save users from typing
__all__ = [
    "Retriever",
    "InMemoryRetriever",
    "RetrievalRequestProcessor",
    "compute_embeddings",
    "write_embeddings",
    "util",
]



# SPDX-License-Identifier: Apache-2.0

"""
Base classes for Input/Output processors
"""

# Standard
import abc
import asyncio
import threading

# Third Party
import aconfig

# Local
from granite_io.backend.base import Backend, ChatCompletionBackend
from granite_io.factory import FactoryConstructible
from granite_io.types import (
    ChatCompletionInputs,
    ChatCompletionResults,
    GenerateInputs,
    GenerateResults,
)


def _workaround_for_horrible_design_flaw_in_asyncio(coroutine_to_run):
    """
    Horrible hack that seems to be the only way to call a coroutine from a non-async
    function that is being called from an async function.

    Creates a background thread with its own event loop. The background thread then
    runs a single task in that event loop.
    """
    result_holder = []

    def _callback():
        # Python's thread library requires you to write your own code to return results
        # or raise exceptions from a background thread.
        try:
            result_holder.append(asyncio.run(coroutine_to_run))
        except Exception as e:  # pylint: disable=W0718
            result_holder.append(e)

    thread = threading.Thread(target=_callback)
    thread.start()
    thread.join()
    if isinstance(result_holder[0], Exception):
        raise result_holder[0]
    return result_holder[0]


class InputOutputProcessor(FactoryConstructible):
    """
    Interface for generic input-output processors. An input-output processor exposes an
    API similar to a chat completions API. Each "completion" call may result in multiple
    LLM calls or even zero LLM calls, depending on the implementation of the
    input-output processor.
    """

    def __init__(self, config: aconfig.Config | None = None):  # pylint: disable=unused-argument
        """By default an IO processor doesn't require config"""

    @abc.abstractmethod
    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        """
        :param inputs: Structured representation of the inputs to a chat completion
            request, possibly including additional fields that only this input-output
            processor can consume

        :returns: The next message that the model produces when fed the specified
            inputs, plus additional information about the low-level request.
        """

    def create_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        """
        Non-async version of :func:`acreate_chat_completion()`

        :param inputs: Structured representation of the inputs to a chat completion
            request, possibly including additional fields that only this input-output
            processor can consume

        :returns: The next message that the model produces when fed the specified
            inputs, plus additional information about the low-level request.
        """
        # Fall back on async version of this method by default.  Subclasses may override
        # this method if they have a more efficient way of doing non-async operation.
        coroutine_to_run = self.acreate_chat_completion(inputs)
        try:  # Exceptions as control flow. Sorry, asyncio forces this design on us.
            asyncio.get_running_loop()

            # If we get here, this code is running inside an async function.
            return _workaround_for_horrible_design_flaw_in_asyncio(coroutine_to_run)
        except RuntimeError:
            # If we get here, this code is not running inside an async function.
            # First we exit the the exception handler; otherwise any exceptions that are
            # thrown from the coroutine will be chained off the current RuntimeError.
            pass
        return asyncio.run(coroutine_to_run)


class ModelDirectInputOutputProcessor(InputOutputProcessor):
    """
    Base class for input-output processors that work by translating inputs to strings
    suitable for tokenization, then passing these strings to a particular model's
    low-level completions API, then parsing raw output strings.
    """

    _backend: Backend | ChatCompletionBackend | None
    """Handle on the current inference engine, required if this io processor's
    :func:`create_chat_completion()` method is going to be used"""

    def __init__(
        self, config: aconfig.Config | None = None, backend: Backend | None = None
    ):
        """
        :param config: Setup config for this IO processor
        :param backend: Handle on inference engine, required if this io processor's
            :func:`create_chat_completion()` method is going to be used
        """
        super().__init__(config)
        self._backend = backend

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        if self._backend is None:
            raise ValueError(
                "Attempted to call create_chat_completion() without "
                "configuring an inference backend."
            )

        # Do not modify `inputs` in place. It is VERY VERY IMPORTANT not to do so.
        # The caller may be using that object for something else.
        generate_inputs = (
            inputs.generate_inputs.model_copy()
            if inputs.generate_inputs is not None
            else GenerateInputs()
        )
        generate_inputs.prompt = self.inputs_to_string(inputs)
        model_output = await self._backend.pipeline(generate_inputs)
        return self.output_to_result(output=model_output, inputs=inputs)

    @abc.abstractmethod
    def inputs_to_string(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> str:
        """
        Convert the structured representation of the inputs to a completion request into
        the string representation of the tokens that should be sent to the model to
        implement said request.

        :param inputs: Structured representation of the inputs
        :param add_generation_prompt: If ``True``, the returned prompt string will
            contain a prefix of the next assistant response for use as a prompt to a
            generation request. Otherwise, the prompt will only contain the messages and
            documents in ``input``.

        :returns: String that can be passed to the model's tokenizer to create a prompt
            for generation.
        """

    @abc.abstractmethod
    def output_to_result(
        self, output: GenerateResults, inputs: ChatCompletionInputs | None = None
    ) -> ChatCompletionResults:
        """
        Convert the structured representation of the inputs to a completion request into
        the string representation of the tokens that should be sent to the model to
        implement said request.

        :param output: Output of the a generation request, potentially incomplete
           if it was a streaming request
        :param inputs: Optional reference to the inputs that caused the model to produce
           ``output``, for validating the correctness of the output. If no inputs are
           provided, this method may skip some validations but will still produce the
           same result.

        :returns: The parsed output so far
        """


class ModelDirectInputOutputProcessorWithGenerate(InputOutputProcessor):
    """
    Base class for IO processors that directly invoke a model and that modify the
    generation arguments of requests, such as by enabling constrained decoding.
    """

    _backend: Backend | None
    """Handle on the current inference engine, required if this io processor's
    :func:`acreate_chat_completion()` method is going to be used"""

    def __init__(
        self, config: aconfig.Config | None = None, backend: Backend | None = None
    ):
        """
        :param config: Setup config for this IO processor
        :param backend: Handle on inference engine, required if this io processor's
            :func:`create_chat_completion()` method is going to be used
        """
        super().__init__(config)
        self._backend = backend

    @abc.abstractmethod
    def inputs_to_generate_inputs(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> GenerateInputs:
        """
        Determine the best generation parameters (including prompt) to pass to the
        backend when running the specified chat completion.

        :param inputs: Structured representation of the inputs
        :param add_generation_prompt: If true, the generated prompt will include the
         appropriate response prefix to trigger generation.

        :returns: A copy of inputs.generate_inputs with appropriate modifications for
         the target model and with the ``prompt`` field populated with an appropriate
         prompt
        """

    @abc.abstractmethod
    def output_to_result(
        self, output: GenerateResults, inputs: ChatCompletionInputs | None = None
    ) -> ChatCompletionResults:
        """
        Convert the structured representation of the inputs to a completion request into
        the string representation of the tokens that should be sent to the model to
        implement said request.

        :param output: Output of the a generation request, potentially incomplete
           if it was a streaming request
        :param inputs: Optional reference to the inputs that caused the model to produce
           ``output``, for validating the correctness of the output. If no inputs are
           provided, this method may skip some validations but will still produce the
           same result.

        :returns: The parsed output so far
        """

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        if self._backend is None:
            raise ValueError(
                "Attempted to call acreate_chat_completion() without "
                "configuring an inference backend."
            )

        # Copy the inputs to prevent the caller experiencing surprise modifications to a
        # local variable.
        inputs = inputs.model_copy()
        generate_inputs = self.inputs_to_generate_inputs(inputs)
        model_output = await self._backend.pipeline(generate_inputs)
        return self.output_to_result(output=model_output, inputs=inputs)


class InputProcessor(FactoryConstructible):
    """
    Interface for generic input processors. An input processor exposes an
    API to transform model completion request into a string representation.
    """

    def __init__(self, config: aconfig.Config | None = None):  # pylint: disable=unused-argument
        """By default an input processor doesn't require config"""

    @abc.abstractmethod
    def transform(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> str:
        """
        Convert the structured representation of the inputs to a completion request into
        the string representation of the tokens that should be sent to the model to
        implement said request.

        :param inputs: Structured representation of the inputs
        :param add_generation_prompt: If ``True``, the returned prompt string will
            contain a prefix of the next assistant response for use as a prompt to a
            generation request. Otherwise, the prompt will only contain the messages and
            documents in ``input``.

        :returns: String that can be passed to the model's tokenizer to create a prompt
            for generation.
        """


class OutputProcessor(FactoryConstructible):
    """
    Interface for generic output processors. An putput processor exposes an
    API to transform model output into a structured representation of the
    information.
    """

    def __init__(self, config: aconfig.Config | None = None):  # pylint: disable=unused-argument
        """By default an output processor doesn't require config"""

    @abc.abstractmethod
    def transform(
        self, output: GenerateResults, inputs: ChatCompletionInputs | None = None
    ) -> ChatCompletionResults:
        """
        Convert the model output generated into a structured representation of the
        information.

        :param output: Output of the a generation request, potentially incomplete
           if it was a streaming request
        :param inputs: Optional reference to the inputs that caused the model to produce
           ``output``, for validating the correctness of the output. If no inputs are
           provided, this method may skip some validations but will still produce the
           same result.

        :returns: The parsed output so far
        """


def make_new_io_processor(
    input_processor: InputProcessor,
    output_processor: OutputProcessor,
    config: aconfig.Config = None,
    backend: Backend | None = None,
) -> ModelDirectInputOutputProcessor:
    """
    Wrapper function that creates an instance of an InputOutputProcessor based on the
    InputProcessor and OutputProcessor passed in the function call.

    :param input_processor: Processor that performs processing of input to model
    :param input_processor: Processor that performs processing of output from the model
    :param config: Setup config for this IO processor
    :param backend: Handle on inference engine, required if this io processor's
        :func:`create_chat_completion()` method is going to be used

    :returns: The IO processor

    :raise: ValueError - If input or output processor is None
    """

    class _InputOutputProcessor(ModelDirectInputOutputProcessor):
        """
        InputOutputProcessor template.

        This InputOutputProcessor is based on the input and the output processors
        passed during creation.
        """

        def __init__(
            self,
            input_processor: InputProcessor,
            output_processor: OutputProcessor,
            config: aconfig.Config = None,
            backend: Backend | None = None,
        ):
            """
            :param input_processor: Processor that performs processing of input to
                model
            :param input_processor: Processor that performs processing of output from
                the model
            :param config: Setup config for this IO processor
            :param backend: Handle on inference engine, required if this io processor's
                :func:`create_chat_completion()` method is going to be used

            :raise: ValueError - If input or output processor is None
            """
            super().__init__(config=config, backend=backend)

            self._input_processor = input_processor
            self._output_processor = output_processor

            if self._input_processor is None:
                raise ValueError(
                    "Attempted to create IO Processor without "
                    "setting an Input Processor ."
                )
            if self._output_processor is None:
                raise ValueError(
                    "Attempted to create IO Processor without "
                    "setting an Output Processor ."
                )

        def inputs_to_string(
            self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
        ) -> str:
            return self._input_processor.transform(inputs, add_generation_prompt)

        def output_to_result(
            self,
            output: GenerateResults,
            inputs: ChatCompletionInputs | None = None,
        ) -> ChatCompletionResults:
            return self._output_processor.transform(output, inputs)

    return _InputOutputProcessor(
        input_processor=input_processor,
        output_processor=output_processor,
        config=config,
        backend=backend,
    )


class RequestProcessor(abc.ABC):
    """
    Base class for objects that modify a ``ChatCompletionInputs`` object in various
    ways, such as by adding RAG documents or changing the content of existing turns.
    """

    @abc.abstractmethod
    async def aprocess(
        self, inputs: ChatCompletionInputs
    ) -> list[ChatCompletionInputs]:
        """
        Subclasses must implement this entry point. Modify the request, potentially
        spawning asynchronous background tasks to perform the modifications.

        :param inputs: The original request

        :returns: One or more modified COPIES of the original request
        """

    def process(self, inputs: ChatCompletionInputs) -> list[ChatCompletionInputs]:
        """
        Subclasses may optionally implement this entry point if they have a more
        efficient way to do non-async operation. The default implementation calls to
        :func:`aprocess()`.

        :param inputs: The original request

        :returns: One or more modified COPIES of the original request
        """
        # Fall back on async version of this method by default.  Subclasses may override
        # this method if they have a more efficient way of doing non-async operation.
        coroutine_to_run = self.aprocess(inputs)
        try:  # Exceptions as control flow. Sorry, asyncio forces this design on us.
            asyncio.get_running_loop()

            # If we get here, this code is running inside an async function.
            return _workaround_for_horrible_design_flaw_in_asyncio(coroutine_to_run)
        except RuntimeError:
            # If we get here, this code is not running inside an async function.
            # First we exit the the exception handler; otherwise any exceptions that are
            # thrown from the coroutine will be chained off the current RuntimeError.
            pass
        return asyncio.run(coroutine_to_run)


class RewriteRequestProcessor(RequestProcessor):
    """
    Request processor that rewrites the last message of a chat completion request by
    passing that message through an IO processor that rewrites or augments messages.
    """

    def __init__(self, io_proc: InputOutputProcessor, top_k: int = 1):
        """
        :param io_proc: IO processor for a model that rewrites the last message in its
         input. Can be a composite IO processor.
        :param top_k: Number of different rewrites to generate
        """
        self._io_proc = io_proc
        self._top_k = top_k

    async def aprocess(
        self, inputs: ChatCompletionInputs
    ) -> list[ChatCompletionInputs]:
        # Generate one or more versions of the last turn.
        new_last_turns = await self._io_proc.acreate_chat_completion(inputs)

        final_results = []
        for result in new_last_turns.results:
            new_messages = inputs.messages.copy()
            new_messages[-1] = inputs.messages[-1].model_copy(
                update={"content": result.next_message.content}
            )
            final_results.append(inputs.model_copy(update={"messages": new_messages}))
        return final_results



# SPDX-License-Identifier: Apache-2.0

"""
This factory utility provides common factory construction semantics and a common
base class for classes that can be constructed via config

Derived from: https://github.com/caikit/caikit/blob/main/caikit/core/toolkit/factory.py
"""

# Standard
from functools import partial, wraps
from typing import Protocol
import abc
import copy
import importlib
import inspect

# Third Party
import aconfig
import jsonschema
import jsonschema.exceptions


class FactoryConstructible(abc.ABC):
    """A class can be constructed by a factory if its constructor takes exactly
    one argument that is an aconfig.Config object and it has a name to identify
    itself with the factory.
    """

    # This is the set of default values for the config that me be set by child
    # implementations
    config_defaults = {}

    # This is the jsonschema dict for this instance's config
    config_schema = {}

    @abc.abstractmethod
    def __init__(self, config: aconfig.Config, **kwargs):
        """A FactoryConstructible object must be constructed with a config
        object that it uses to pull in all configuration
        """


class FactoryCallable(Protocol):
    """A callable that can be stored and constructed by a factory"""

    def __call__(self, config: aconfig.Config, **kwargs):
        """Factory callables must take config as the first argument"""


class Factory:
    """The base factory class implements all common factory functionality to
    read a designated portion of config and instantiate an instance of the
    registered classes.
    """

    # The keys in the instance config
    TYPE_KEY = "type"
    CONFIG_KEY = "config"

    def __init__(self, name: str):
        """Construct with the path in the global config where this factory's
        configuration lives.
        """
        self._name = name
        self._registered_types = {}

    @property
    def name(self) -> str:
        return self._name

    def register(
        self, constructible: type[FactoryConstructible] | FactoryCallable, *aliases
    ):
        """Register the given constructible/callable"""
        base_name = getattr(constructible, "name", None) or constructible.__name__
        names = [base_name] + list(aliases)
        assert not any(name in self._registered_types for name in names), (
            f"Conflicting registration of {base_name} (aliases: {aliases})"
        )
        for name in names:
            self._registered_types[name] = constructible

    def registered_types(self) -> list[str]:
        """Get the list of registered types"""
        return list(sorted(self._registered_types.keys()))

    @classmethod
    def _get_construct_instance_config(
        cls,
        arg_one: str | dict,
        instance_config: dict | None = None,
    ) -> dict:
        """Shared helper for parsing args in construct"""
        if isinstance(arg_one, str):
            inst_type = arg_one
            cfg = instance_config or {}
            instance_config = {
                cls.TYPE_KEY: inst_type,
                cls.CONFIG_KEY: cfg,
            }
        elif isinstance(arg_one, dict):
            assert instance_config is None, (
                "Cannot pass config as both first and second argument"
            )
            instance_config = arg_one
        else:
            raise TypeError(f"Invalid argument type {type(arg_one)} for construct")
        return instance_config

    def construct(
        self,
        arg_one: str | dict,
        instance_config: dict | None = None,
        *,
        validate: bool = True,
        **kwargs,
    ) -> FactoryConstructible:
        """Construct an instance of the given type"""
        instance_config = self._get_construct_instance_config(arg_one, instance_config)
        inst_type = instance_config.get(self.__class__.TYPE_KEY)
        cfg = instance_config.get(self.__class__.CONFIG_KEY)
        inst_cls = self._registered_types.get(inst_type)
        if inst_cls is None:
            raise ValueError(f"No {self.name} class registered for type {inst_type}")
        inst_cfg = aconfig.Config(
            _merge_configs(
                copy.deepcopy(inst_cls.config_defaults),
                cfg,
            ),
            override_env_vars=True,
        )
        if validate:
            # NOTE: This explicitly allows additional properties
            jsonschema.validate(instance=inst_cfg, schema=inst_cls.config_schema)
        return inst_cls(inst_cfg, **kwargs)

    def decorator(
        self,
        name: str,
        *aliases: list[str],
        config_schema: dict | None = None,
        config_defaults: dict | None = None,
    ) -> callable:
        """Decorator to register a class or function with the factory"""

        def _decorator(cls_or_fn, config_schema, config_defaults):
            if inspect.isfunction(cls_or_fn):
                sig = inspect.signature(cls_or_fn)

                # If no schema given, deduce it from the type signature
                if config_schema is None:
                    config_schema = _get_jsonschema(
                        {k: v.annotation for k, v in sig.parameters.items()}
                    )
                if config_defaults is None:
                    config_defaults = {
                        k: p.default
                        for k, p in sig.parameters.items()
                        if p.default is not inspect._empty
                    }

                # Make sure the schema matches the kwargs in the function
                schema_properties = (config_schema or {}).get("properties", {})
                if not all(prop in sig.parameters for prop in schema_properties):
                    raise ValueError(
                        f"Schema properties {schema_properties} do not match "
                        f"function kwargs {sig.parameters}"
                    )
                sig_kwargs = list(sig.parameters.keys())

                @wraps(cls_or_fn)
                def _wrapper_callable(config: aconfig.Config, **kwargs):
                    config_kwargs = {
                        k: v
                        for k, v in (config or aconfig.Config({})).items()
                        if k in sig_kwargs
                    }
                    return partial(cls_or_fn, **config_kwargs, **kwargs)

                cls = _wrapper_callable

            elif inspect.isclass(cls_or_fn):
                cls = cls_or_fn

            else:
                raise TypeError("Decorator must wrap a class or function")

            # Add the class attributes
            cls.name = name
            if config_schema is not None:
                cls.config_schema = config_schema
            if config_defaults is not None:
                cls.config_defaults = config_defaults

            # Register with the factory
            self.register(cls, *aliases)

            # Return the wrapped object (class or function)
            return cls_or_fn

        return partial(
            _decorator, config_schema=config_schema, config_defaults=config_defaults
        )


class ImportableFactory(Factory):
    """An ImportableFactory extends the base Factory to allow the construction
    to specify an "import_class" field that will be used to import and register
    the implementation class before attempting to initialize it.
    """

    IMPORT_CLASS_KEY = "import_class"

    def construct(
        self,
        arg_one: str | dict,
        instance_config: dict | None = None,
        **kwargs,
    ):
        # Look for an import_class and import and register it if found
        instance_config = self._get_construct_instance_config(arg_one, instance_config)
        import_class_val = instance_config.get(self.__class__.IMPORT_CLASS_KEY)
        if import_class_val:
            assert isinstance(import_class_val, str)
            module_name, class_name = import_class_val.rsplit(".", 1)
            imported_module = importlib.import_module(module_name)
            imported_class = getattr(imported_module, class_name)
            self.register(imported_class)
        return super().construct(instance_config, **kwargs)


## Implementation Details ######################################################


_CONFIG_TYPE = dict | aconfig.Config


def _merge_configs(
    base: _CONFIG_TYPE | None,
    overrides: _CONFIG_TYPE | None,
) -> aconfig.Config:
    """Helper to perform a deep merge of the overrides into the base. The merge
    is done in place, but the resulting dict is also returned for convenience.
    The merge logic is quite simple: If both the base and overrides have a key
    and the type of the key for both is a dict, recursively merge, otherwise
    set the base value to the override value.
    Args:
        base: The base config that will be updated with the overrides
        overrides: The override config
    Returns:
        merged:
            The merged results of overrides merged onto base
    """
    # Handle none args
    if base is None:
        return overrides or {}
    if overrides is None:
        return base or {}

    # Do the deep merge
    for key, value in overrides.items():
        if (
            key not in base
            or not isinstance(base[key], dict)
            or not isinstance(value, dict)
        ):
            base[key] = value
        else:
            base[key] = _merge_configs(base[key], value)

    return base


def _get_jsonschema(sig_types: dict[str:type]) -> dict:
    """Given a mapping from keywords to type hints, create a jsonschema"""
    if None in sig_types.values() or inspect._empty in sig_types.values():
        raise ValueError("Cannot deduce config schema from function without type hints")
    all_sig_types = set(sig_types.values())
    validator = jsonschema.validators.validator_for({})
    type_mapping = {
        t: x[0] if x else None
        for t, x in zip(
            all_sig_types,
            [
                [
                    y
                    for y in validator.TYPE_CHECKER._type_checkers
                    if validator.TYPE_CHECKER.is_type(t(), y)
                ]
                for t in all_sig_types
            ],
            strict=True,
        )
    }
    return {"properties": {k: {"type": type_mapping[v]} for k, v in sig_types.items()}}



# SPDX-License-Identifier: Apache-2.0

# Third Party
import aconfig

# Local
from granite_io.backend.base import Backend
from granite_io.io.base import ModelDirectInputOutputProcessor
from granite_io.io.consts import (
    _GRANITE_3_2_2B_HF,
    _GRANITE_3_2_2B_OLLAMA,
    _GRANITE_3_2_MODEL_NAME,
)
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2InputProcessor,
)
from granite_io.io.granite_3_2.output_processors.granite_3_2_output_processor import (
    Granite3Point2OutputProcessor,
)
from granite_io.io.registry import io_processor
from granite_io.types import (
    ChatCompletionInputs,
    ChatCompletionResults,
    GenerateResults,
)


@io_processor(
    _GRANITE_3_2_MODEL_NAME,
    # Huggingface
    _GRANITE_3_2_2B_HF,
    "ibm-granite/granite-3.2-8b-instruct",
    # Ollama
    "granite3.2",
    "granite3.2:8b",
    _GRANITE_3_2_2B_OLLAMA,
    # RITS
    "ibm-granite/granite-8b-instruct-preview-4k",
)
class Granite3Point2InputOutputProcessor(ModelDirectInputOutputProcessor):
    """
    Input/output processor for version 3.2 of the main Granite models, all sizes.

    This input/output processor is based on the Jinja template that was used during
    supervised fine tuning of these models.
    """

    def __init__(self, config: aconfig.Config = None, backend: Backend | None = None):
        """
        :param backend: Handle on inference engine, required if this io processor's
            :func:`create_chat_completion()` method is going to be used
        """
        super().__init__(backend=backend)

    def inputs_to_string(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> str:
        input_processor = Granite3Point2InputProcessor()
        return input_processor.transform(inputs, add_generation_prompt)

    def output_to_result(
        self,
        output: GenerateResults,
        inputs: ChatCompletionInputs | None = None,
    ) -> ChatCompletionResults:
        output_processor = Granite3Point2OutputProcessor()
        return output_processor.transform(output, inputs)



# SPDX-License-Identifier: Apache-2.0

"""
This example show how to use sequential inference scaling
to enforce the that model response has no hallucination.
"""

# Third Party
from granite_io_pdl.sequential_scaling import SequentialScalingInputOutputProcessor

# Local
from granite_io.types import ChatCompletionInputs, ChatCompletionResults, UserMessage


def hallucination_validator(processor_outputs: ChatCompletionResults):
    return not processor_outputs.results[0].next_message.hallucinations


model_name = "granite3.2:8b"
io_processor = SequentialScalingInputOutputProcessor(
    model=model_name, backend="openai", validator=hallucination_validator
)
question = "What is RAG?"
messages = [UserMessage(content=question)]

documents = [
    {
        "text": "Dynamic Data: Automatically generate and refine graphs to best fit your domain and ontology needs."  # noqa: E501
    },
    {
        "text": "RAG, retrieval-augmented generation, is a technique that grants generative artificial intelligence models information retrieval capabilities."  # noqa: E501
    },
]

outputs = io_processor.create_chat_completion(
    ChatCompletionInputs(
        messages=messages,
        documents=documents,
        controls={"hallucinations": True},
    )
)
print("\n\n>> Response:\n")
response = outputs.results[0].next_message.content
print(response)

print("\n\n>> Hallucinations:\n")
if outputs.results[0].next_message.hallucinations:
    print("We are sorry, we tried to not hallucinate!")
else:
    print("We are happy to tell you that we did not hallucinate!")



# SPDX-License-Identifier: Apache-2.0

"""
Base class for backends
"""

# Standard
import abc

# Third Party
import aconfig

# Local
from granite_io.factory import FactoryConstructible
from granite_io.types import (
    ChatCompletionInputs,
    GenerateInputs,
    GenerateResult,
    GenerateResults,
)


# We put this base class here to avoid circular imports
class Backend(FactoryConstructible):
    """
    Base class for classes that provide an interface to a string-based completions API
    for a model.

    This base class exists for two reasons: It smoothes out the differences between
    APIs, and it provides a level of indirection so that backend inference libraries can
    be optional dependencies of this Python package.
    """

    _model_str: str

    def __init__(self, config: aconfig.Config):
        self._model_str = config.model_name

    async def __call__(self, inputs: GenerateInputs) -> GenerateResults:
        return await self.pipeline(inputs)

    async def pipeline(self, inputs: GenerateInputs) -> GenerateResults:
        """
        Process input, call completion (generate), process and return output
        """
        inputs = self.process_input(inputs)
        output = await self.generate(inputs)
        return self.process_output(output)

    def process_input(self, inputs: GenerateInputs) -> GenerateInputs:
        """
        Process inputs to prepare them for completion.create() (or generate())

        Args:
            inputs (GenerateInputs): the inputs which are being processed
             to be used in target backend call.

        Returns:
            GenerateInputs: a new copy of inputs, modified for the Backend
        """

        # Do not modify `inputs` in place.
        inputs_copy = inputs.model_copy() if inputs is not None else GenerateInputs()

        # Add required model, if missing
        if not inputs_copy.model:
            inputs_copy.model = self._model_str

        # n (a.k.a. num_return_sequences) validation
        n = inputs_copy.n
        if n is not None:  # noqa SIM102
            if n < 1:
                raise ValueError(f"Invalid value for n ({n})")
            if n > 1:
                # best_of must be >= n
                best_of = inputs_copy.best_of
                if best_of is None or best_of < n:
                    inputs_copy.best_of = n

        # Some backends prefer an array to a string
        if isinstance(inputs_copy.stop, str):
            inputs_copy.stop = [inputs_copy.stop]

        return inputs_copy

    @abc.abstractmethod
    async def generate(self, inputs: GenerateInputs) -> GenerateResults:
        """
        Callback to invoke the model to generate a response.
        """
        raise NotImplementedError()

    def process_output(self, outputs):
        """
        Process output from completion.create() (or generate())
        """
        results = []
        for choice in outputs.choices:
            results.append(
                GenerateResult(
                    completion_string=choice.text,
                    completion_tokens=[],  # Not part of the OpenAI spec
                    stop_reason=choice.finish_reason,
                )
            )

        return GenerateResults(results=results)

    @staticmethod
    def kwarg_alias(kw_dict, preferred_key, alias_key):
        """Migrate kwarg k,v from alias key to preferred key"""
        if alias_key in kw_dict:
            alias_value = kw_dict.get(alias_key)
            # Alias is removed
            del kw_dict[alias_key]
            # Preferred key/value gets priority if both are specified
            # TODO: optionally error on redundant kwargs(?)
            if preferred_key not in kw_dict:
                # Alias value is set to preferred key (iff not already set)
                kw_dict[preferred_key] = alias_value
        return kw_dict  # return reference for convenience


class ChatCompletionBackend(FactoryConstructible):
    """
    Base class for classes that provide an interface to a chat completions API.

    This base class exists for two reasons: It smooths out the differences between
    APIs, and it provides a level of indirection so that backend inference libraries can
    be optional dependencies of this Python package.
    """

    @abc.abstractmethod
    async def create_chat_completion(self, input_chat: ChatCompletionInputs) -> str:
        """
        Callback to invoke the model to generate a response.
        """
        raise NotImplementedError()



# SPDX-License-Identifier: Apache-2.0

"""
Test cases for query_rewrite.py
"""

# Standard
import datetime
import textwrap

# Third Party
import pytest

# Local
from granite_io.backend.vllm_server import LocalVLLMServer
from granite_io.io.base import RewriteRequestProcessor
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Granite3Point2Inputs,
    override_date_for_testing,
)
from granite_io.io.query_rewrite import QueryRewriteIOProcessor
from granite_io.types import (
    GenerateResult,
    GenerateResults,
)

_EXAMPLE_CHAT_INPUT = Granite3Point2Inputs.model_validate(
    {
        "messages": [
            {"role": "assistant", "content": "Welcome to pet questions!"},
            {
                "role": "user",
                "content": "I have two pets, a dog named Rex and a cat named Lucy.",
            },
            {
                "role": "assistant",
                "content": "Great, what would you like to share about them?",
            },
            {
                "role": "user",
                "content": "Rex spends a lot of time in the backyard and outdoors, "
                "and Luna is always inside.",
            },
            {
                "role": "assistant",
                "content": "Sounds good! Rex must love exploring outside, while Lucy "
                "probably enjoys her cozy indoor life.",
            },
            {
                "role": "user",
                "content": "But is he more likely to get fleas because of that?",
            },
        ],
        "generate_inputs": {
            "temperature": 0.0  # Ensure consistency across runs
        },
    }
)


def _make_result(content: str):
    """Convenience method to create a fake model output object."""
    return GenerateResult(
        completion_string=content, completion_tokens=[], stop_reason="dummy stop reason"
    )


_TODAYS_DATE = datetime.datetime.now().strftime("%B %d, %Y")

INSTRUCTION_TEXT = (
    "Reword the final utterance from the USER into a single utterance that doesn't "
    "need the prior conversation history to understand the user's intent. If the final "
    "utterance is a clear and standalone question, please DO NOT attempt to rewrite "
    "it, rather output the last user utterance as is. "
)
JSON = 'Your output format should be in JSON: { "rewritten_question": <REWRITE> }'
REWRITE_PROMPT = (
    "<|start_of_role|>rewrite: " + INSTRUCTION_TEXT + JSON + "<|end_of_role|>"
)


def test_canned_input():
    """
    Validate that the I/O processor handles a single instance of canned input in the
    expected way.
    """
    io_processor = QueryRewriteIOProcessor(None)
    output = io_processor.inputs_to_generate_inputs(_EXAMPLE_CHAT_INPUT)

    expected_output = textwrap.dedent(f"""<|start_of_role|>system<|end_of_role|>Knowledge Cutoff Date: April 2024.
Today's Date: {_TODAYS_DATE}.
You are Granite, developed by IBM. You are a helpful AI assistant.<|end_of_text|>
<|start_of_role|>assistant<|end_of_role|>Welcome to pet questions!<|end_of_text|>
<|start_of_role|>user<|end_of_role|>I have two pets, a dog named Rex and a cat named Lucy.<|end_of_text|>
<|start_of_role|>assistant<|end_of_role|>Great, what would you like to share about them?<|end_of_text|>
<|start_of_role|>user<|end_of_role|>Rex spends a lot of time in the backyard and outdoors, and Luna is always inside.<|end_of_text|>
<|start_of_role|>assistant<|end_of_role|>Sounds good! Rex must love exploring outside, while Lucy probably enjoys her cozy indoor life.<|end_of_text|>
<|start_of_role|>user<|end_of_role|>But is he more likely to get fleas because of that?<|end_of_text|>
{REWRITE_PROMPT}""")  # noqa: E501

    print(f"---\n|{output.prompt}|\n---\n")
    print(f"|{expected_output}|\n---")

    assert output.prompt == expected_output


def test_canned_output():
    """
    Validate that the I/O processor handles a single instance of canned model output
    in the expected way.
    """
    io_processor = QueryRewriteIOProcessor(None)

    raw_output_to_expected = [
        (
            '{ "rewritten_question": "Is Rex more likely to get fleas because he spends a lot of time outdoors?" }',  # noqa: E501
            "Is Rex more likely to get fleas because he spends a lot of time outdoors?",
        ),
        # Current code passes through malformed data unchanged
        ("<invalid model response>", "<invalid model response>"),
    ]

    # Single output
    for raw_output, expected in raw_output_to_expected:
        output = io_processor.output_to_result(
            GenerateResults(results=[_make_result(raw_output)]), _EXAMPLE_CHAT_INPUT
        )
        print(f"\n{output}\n")

        assert len(output.results) == 1
        assert output.results[0].next_message.content == expected

    # Multiple outputs
    multi_raw_output = [
        _make_result(raw_output) for raw_output, _ in raw_output_to_expected
    ]
    multi_expected = [expected for _, expected in raw_output_to_expected]
    multi_output = io_processor.output_to_result(
        GenerateResults(results=multi_raw_output), _EXAMPLE_CHAT_INPUT
    )
    multi_output_strs = [r.next_message.content for r in multi_output.results]
    print(f"\n{output}\n")
    assert multi_output_strs == multi_expected


@pytest.mark.vcr
def test_run_model(lora_server: LocalVLLMServer, fake_date: str):
    """
    Run a chat completion through the LoRA adapter using the I/O processor.
    """
    backend = lora_server.make_lora_backend("query_rewrite")
    io_proc = QueryRewriteIOProcessor(backend)
    override_date_for_testing(fake_date)  # For consistent VCR output

    # Pass our example input through the I/O processor and retrieve the result
    chat_result = io_proc.create_chat_completion(_EXAMPLE_CHAT_INPUT)

    print(f"\n{chat_result.results[0].next_message.content}\n")
    assert (
        chat_result.results[0].next_message.content
        == "Is Rex more likely to get fleas because he spends a lot of time outdoors?"
    )


@pytest.mark.vcr(record_mode="new_episodes")
def test_request_processor(lora_server: LocalVLLMServer, fake_date: str):
    """
    Run a chat completion through the LoRA adapter using a RequestProcessor
    """
    backend = lora_server.make_lora_backend("query_rewrite")
    io_proc = QueryRewriteIOProcessor(backend)
    request_proc = RewriteRequestProcessor(io_proc)
    override_date_for_testing(fake_date)  # For consistent VCR output

    # Pass our example input through the rewrite and retrieve the result
    rewrites = request_proc.process(_EXAMPLE_CHAT_INPUT)
    assert len(rewrites) == 1
    rewritten_input = rewrites[0]

    assert len(rewritten_input.messages) == len(_EXAMPLE_CHAT_INPUT.messages)
    assert (
        rewritten_input.messages[-1].content
        == "Is Rex more likely to get fleas because he spends a lot of time outdoors?"
    )



# SPDX-License-Identifier: Apache-2.0


# Local
from .answerability import AnswerabilityCompositeIOProcessor, AnswerabilityIOProcessor

# Expose public symbols at `granite_io.io.certainty` to save users from typing
__all__ = ["AnswerabilityIOProcessor", "AnswerabilityCompositeIOProcessor"]



# SPDX-License-Identifier: Apache-2.0

# Third Party
import aconfig

# Local
from granite_io.backend.base import Backend
from granite_io.io.base import ModelDirectInputOutputProcessor
from granite_io.io.consts import (
    _GRANITE_3_3_2B_HF,
    _GRANITE_3_3_2B_OLLAMA,
    _GRANITE_3_3_8B_HF,
    _GRANITE_3_3_8B_OLLAMA,
    _GRANITE_3_3_MODEL_NAME,
)
from granite_io.io.granite_3_3.input_processors.granite_3_3_input_processor import (
    Granite3Point3InputProcessor,
)
from granite_io.io.granite_3_3.output_processors.granite_3_3_output_processor import (
    Granite3Point3OutputProcessor,
)
from granite_io.io.registry import io_processor
from granite_io.types import (
    ChatCompletionInputs,
    ChatCompletionResults,
    GenerateInputs,
    GenerateResults,
)


@io_processor(
    _GRANITE_3_3_MODEL_NAME,
    # Huggingface
    _GRANITE_3_3_2B_HF,
    _GRANITE_3_3_8B_HF,
    # Ollama
    "granite3.3",
    _GRANITE_3_3_2B_OLLAMA,
    _GRANITE_3_3_8B_OLLAMA,
)
class Granite3Point3InputOutputProcessor(ModelDirectInputOutputProcessor):
    """
    Input/output processor for version 3.3 of the main Granite models, all sizes.

    This input/output processor is based on the Jinja template that was used during
    supervised fine-tuning of these models.
    """

    def __init__(self, config: aconfig.Config = None, backend: Backend | None = None):
        """
        :param backend: Handle on inference engine, required if this io processor's
            :func:`create_chat_completion()` method is going to be used
        """
        super().__init__(backend=backend)

    async def acreate_chat_completion(
        self, inputs: ChatCompletionInputs
    ) -> ChatCompletionResults:
        if self._backend is None:
            raise ValueError(
                "Attempted to call create_chat_completion() without "
                "configuring an inference backend."
            )

        # Do not modify `inputs` in place. It is VERY VERY IMPORTANT not to do so.
        # The caller may be using that object for something else.
        generate_inputs = (
            inputs.generate_inputs.model_copy()
            if inputs.generate_inputs is not None
            else GenerateInputs()
        )
        generate_inputs.prompt = self.inputs_to_string(inputs)

        # Add skip_special_tokens = False to see citations.
        if inputs.controls and inputs.controls.get("citations"):
            extra_body = generate_inputs.extra_body or {}
            extra_body["skip_special_tokens"] = False
            generate_inputs.extra_body = extra_body
        model_output = await self._backend.pipeline(generate_inputs)
        return self.output_to_result(output=model_output, inputs=inputs)

    def inputs_to_string(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> str:
        input_processor = Granite3Point3InputProcessor()
        return input_processor.transform(inputs, add_generation_prompt)

    def output_to_result(
        self,
        output: GenerateResults,
        inputs: ChatCompletionInputs | None = None,
    ) -> ChatCompletionResults:
        output_processor = Granite3Point3OutputProcessor()
        return output_processor.transform(output, inputs)



# SPDX-License-Identifier: Apache-2.0

"""
This example shows how to infer or call a model using the framework and an OpenAI
backend to serve the model.

It uses MBRD and ROUGE scoring for majority voting to decide on best answer to use
from number of model sample outputs.
"""

# Local
from granite_io import make_backend, make_io_processor
from granite_io.io.voting import MBRDMajorityVotingProcessor
from granite_io.types import ChatCompletionInputs, UserMessage

# By default the OpenAI backend is configured for Ollama server running locally using
# default url. Ollama howevere does not support multiple completions per
# prompt i.e. "n" param below. You will need to use a backend that supports multiple
# completions to fully see majority voting happening. Backends like vLLM support
# multiple completions. You will need to update url and key when connecting to another
# backend.
model_name = "granite3.2:8b"
# openai_base_url = ""
# openai_api_key = ""
base_processor = make_io_processor(
    model_name,
    backend=make_backend(
        "openai",
        {
            "model_name": model_name,
            # "openai_base_url": openai_base_url,
            # "openai_api_key": openai_api_key,
        },
    ),
)
question = "What is 234651 + 13425?\nAnswer with just a number please."
messages = [UserMessage(content=question)]
completion_inputs = ChatCompletionInputs(
    messages=messages,
    thinking=True,
    generate_inputs={"n": 20, "temperature": 0.6, "max_tokens": 1024},
)
voting_processor = MBRDMajorityVotingProcessor(base_processor)
results = voting_processor.create_chat_completion(completion_inputs)
print("Output from base model augmented with MBRD majority voting:")
# This should be only one output, the majority voted answer
for result_num, r in enumerate(results.results):
    print(f"{result_num + 1}: {r.next_message.content}")

# What's the actual answer?
print(f"---------\nThe actual answer is: {234651 + 13425}")



# SPDX-License-Identifier: Apache-2.0

"""Various utility functions relating to the MTRAG benchmark data set."""

# Standard
import os
import pathlib
import urllib
import urllib.error
import zipfile

# Third Party
import pyarrow as pa
import pyarrow.json as pj


def download_mtrag_corpus(target_dir: str, corpus_name: str) -> pathlib.Path:
    """
    Download a corpus file from the MTRAG benchmark if the file hasn't already present.

    :param target_dir: Location where the file should be written if not already present.
    :param corpus_name: Should be one of "cloud", "clapnq", "fiqa", or "govt"

    :returns: Path to the downloaded (or cached) file.
    """
    corpus_names = ("cloud", "clapnq", "fiqa", "govt")
    if corpus_name not in corpus_names:
        raise ValueError(f"Corpus name must be one of {corpus_names}")

    target_file = pathlib.Path(target_dir) / f"{corpus_name}.jsonl.zip"
    if not os.path.exists(target_file):
        source_url = (
            f"https://github.com/IBM/mt-rag-benchmark/raw/refs/heads/main/"
            f"corpora/{corpus_name}.jsonl.zip"
        )
        urllib.request.urlretrieve(source_url, target_file)
    return target_file


MB = 1048576  # 1MB in bytes


def read_mtrag_corpus(corpus_file: str | pathlib.Path) -> pa.Table:
    """
    Read the documents from one of the MTRAG benchmark's corpora.

    :param corpus_file: Location where the corpus data is located.

    :returns: Documents from the corpus as a Pyarrow table, with schema
        ``["id", "url", "title", "text"]``
    """
    if not isinstance(corpus_file, pathlib.Path):
        corpus_file = pathlib.Path(corpus_file)
    tables = []
    with zipfile.ZipFile(corpus_file, "r") as z:
        for name in z.namelist():
            tables.append(
                # pylint: disable=consider-using-with
                pj.read_json(z.open(name), pj.ReadOptions(block_size=100 * MB))
            )
    t = pa.concat_tables(tables)

    # Normalize schema
    # Guess at ID column
    id_candidates = [n for n in t.schema.names if "id" in n]
    if len(id_candidates) != 1:
        raise TypeError(f"Couldn't guess ID column; candidates are {id_candidates}")
    t = t.rename_columns(["id" if n == id_candidates[0] else n for n in t.schema.names])
    if "text" not in t.schema.names:
        raise TypeError(f"No text column (columns are {t.schema.names})")

    if "title" not in t.schema.names and "url" in t.schema.names:
        # Use URL as title if there is no title
        t = t.append_column("title", t.column("url"))

    # Fill in missing columns
    for name in ("title", "url"):
        if name not in t.schema.names:
            t = t.append_column(name, pa.array([None] * t.num_rows, type=pa.string()))

    # Remove extra columns
    t = t.select(["id", "url", "title", "text"])
    return t


def download_mtrag_embeddings(embedding_name: str, corpus_name: str, target_dir: str):
    """
    Download precomputed embeddings for a corpus in the MTRAG benchmark.

    :param embedding_name: Name of SentenceTransformers embedding model that was used
     to create the embeddings.
    :param corpus_name: Should be one of "cloud", "clapnq", "fiqa", or "govt"
    :param target_dir: Location where Parquet files with names "part_001.parquet",
     "part_002.parquet", etc. will be written.
    """
    corpus_names = ("cloud", "clapnq", "fiqa", "govt")
    if corpus_name not in corpus_names:
        raise ValueError(f"Corpus name must be one of {corpus_names}")

    target_root = pathlib.Path(target_dir) / f"{corpus_name}_{embedding_name}.parquet"
    if not os.path.exists(target_root):
        os.makedirs(target_root)

    part_num = 1
    repo_root = "https://github.com/frreiss/mt-rag-embeddings"
    while True:
        # Download part_001.parquet, part_002.parquet, etc. until a download fails.
        parquet_file_name = f"part_{part_num:03d}.parquet"
        source_url = (
            f"{repo_root}/raw/refs/heads/main/{embedding_name}/{corpus_name}.parquet"
            f"/{parquet_file_name}"
        )
        target_file = target_root / parquet_file_name
        try:
            urllib.request.urlretrieve(source_url, target_file)
            part_num += 1
        except urllib.error.HTTPError:
            # Found all the parts; flow through
            break

    if part_num == 1:
        raise ValueError(
            f"No precomputed embeddings for MTRAG corpus '{corpus_name}' "
            f"with embedding model '{embedding_name}' found at {repo_root}"
        )



# SPDX-License-Identifier: Apache-2.0

"""
Registry of all backends
"""

# Local
from granite_io.factory import ImportableFactory

_BACKEND_FACTORY = ImportableFactory("backends")
backend = _BACKEND_FACTORY.decorator
make_backend = _BACKEND_FACTORY.construct



"""
The voting module holds I/O processors that perform various types of majority voting.
"""

# Local
from granite_io.io.voting.mbrd import MBRDMajorityVotingProcessor  # noqa: F401
from granite_io.io.voting.simple import (  # noqa: F401
    MajorityVotingProcessor,
    integer_normalizer,
)



# SPDX-License-Identifier: Apache-2.0

# Standard
import abc
import os
import pathlib
import shutil

# Third Party
import nltk
import numpy as np
import torch

# Local
from granite_io.io.base import ChatCompletionInputs, RequestProcessor
from granite_io.types import Document


class Retriever(abc.ABC):
    """
    Base class for document retrievers. Provides APIs for searching by text snippet and
    for inserting new documents.
    """

    @abc.abstractmethod
    def retrieve(self, query: str, top_k: int = 10) -> list[Document]:
        """Retrieve the top k matches of a query from the corpus.

        :param query: Query string to use for lookup
        :param top_k: Number of top-k results to return

        :returns: Pyarrow table with fields: "id", "title", "url", "begin", "end",
            "text", "score"
        """


def _sliding_windows(
    token_offsets: list[tuple[int, int]],
    sentence_offsets: list[tuple[int, int]],
    window_size: int,
    overlap: int,
) -> list[tuple[int, int]]:
    """Compute window boundaries for slicing a document into chunks.

    :param token_offsets: Character offsets of the tokens, as begin, end pairs
    :param sentence_offsets: Character offsets of sentence boundaries, as begin, end
    :param window_size: Target maximum size of the windows, in tokens
    :param overlap: How many tokens each window should overlap with next
     and previous

    :returns: Tuples of begin and end offsets of windows
    """
    result = []
    cur_start = 0
    doc_len = len(token_offsets)

    # Create some lookup tables
    sentence_begins = {t[0] for t in sentence_offsets}
    sentence_ends = {t[1] for t in sentence_offsets}

    while len(result) == 0 or cur_start < doc_len - overlap:
        # Compute boundaries between the three regions of the window
        cur_end = min(cur_start + window_size, doc_len)
        begin_overlap_end = min(cur_start + overlap, doc_len)
        end_overlap_begin = min(cur_start + window_size - overlap, cur_end)

        # Advance start of window until we cross a sentence boundary, but don't go past
        # the overlap area.
        cur_start_char = token_offsets[cur_start][0]
        max_start_char = token_offsets[begin_overlap_end - 1][0]
        adjusted_start_char = cur_start_char
        while (
            adjusted_start_char not in sentence_begins
            and adjusted_start_char < max_start_char
        ):
            adjusted_start_char += 1
        if adjusted_start_char == max_start_char:
            adjusted_start_char = cur_start_char
        adjusted_start = cur_start
        while token_offsets[adjusted_start][0] < adjusted_start_char:
            adjusted_start += 1

        # print(f"Adjusted start {cur_start} => {adjusted_start}")

        # Move end of window back until we cross a sentence boundary, but don't go past
        # the overlap area
        cur_end_char = token_offsets[cur_end - 1][1]
        min_end_char = token_offsets[end_overlap_begin - 1][1]
        adjusted_end_char = cur_end_char
        while (
            adjusted_end_char not in sentence_ends and adjusted_end_char > min_end_char
        ):
            adjusted_end_char -= 1
        if adjusted_end_char == min_end_char:
            adjusted_end_char = cur_end_char
        adjusted_end = cur_end
        while token_offsets[adjusted_end - 1][1] > adjusted_end_char:
            adjusted_end -= 1

        # print(f"Adjusted end {cur_end} => {adjusted_end}")

        result.append((adjusted_start, adjusted_end))
        cur_start = adjusted_end - overlap
    return result


def compute_embeddings(
    corpus,  #: pa.Table,
    embedding_model_name: str,
    chunk_size: int = 512,
    overlap: int = 128,
):  # "-> pa.Table:"
    # pylint: disable=too-many-locals
    """
    Split documents into windows and compute embeddings for each of the the windows.

    :param corpus: PyArrow Table of documents as returned by :func:`read_corpus()`.
        Should have the columns ``["id", "url", "title", "text"]``
    :param embedding_model_name: Hugging Face model name for the model that computes
        embeddings. Also used for tokenizing.
    :param chunk_size: Maximum size of chunks to split documents into, in embedding
        model tokens; must be less than or equal to the embedding model's maximum
        sequence length.
    :param overlap: Target overlap between adjacent chunks, in embedding model tokens.
        Actual begins and ends of chunks will be on sentence boundaries.

    :returns: PyArrow Table of chunks of the corpus, with schema
        ````["id", "url", "title", "begin", "end", "text", "embedding"]``
    """
    # Third Party
    import pyarrow as pa
    import sentence_transformers

    embedding_model = sentence_transformers.SentenceTransformer(embedding_model_name)
    sentence_splitter = nltk.tokenize.punkt.PunktSentenceTokenizer()

    # Corpora currently fit in memory, so just iterate with a for loop
    id_list = []
    url_list = []
    title_list = []
    text_list = []
    begin_list = []
    end_list = []

    for row in corpus.to_pylist():
        doc_text = row["text"]
        tokenizer_output = embedding_model.tokenizer(
            doc_text, return_offsets_mapping=True, return_attention_mask=False
        )

        # doc_tokens = tokenizer_output["input_ids"]
        token_offsets = tokenizer_output["offset_mapping"]
        sentence_offsets = list(sentence_splitter.span_tokenize(doc_text))

        # For some reason, the tokenizer likes to give the last token a character
        # range from zero to zero.
        if token_offsets[-1] == (0, 0) and len(token_offsets) > 1:
            last_char = token_offsets[-2][1]
            token_offsets[-1] = (last_char, last_char)

        # Pick sliding windows that start and end on sentence boundaries
        offsets = _sliding_windows(token_offsets, sentence_offsets, chunk_size, overlap)

        for begin_tok, end_tok in offsets:
            # Use the token offsets to find the appropriate ranges of characters
            # for the windows. Note that we cannnot do this by detokenizing, as the
            # tokenizer may be case-insensitive.
            begin_char = token_offsets[begin_tok][0]
            end_char = token_offsets[end_tok - 1][1]

            id_list.append(row["id"])
            url_list.append(row["url"])
            title_list.append(row["title"])
            text_list.append(doc_text[begin_char:end_char])
            begin_list.append(begin_char)
            end_list.append(end_char)
    result = pa.table(
        {
            "id": pa.array(id_list),
            "url": pa.array(url_list),
            "title": pa.array(title_list),
            "begin": pa.array(begin_list),
            "end": pa.array(end_list),
            "text": pa.array(text_list),
        }
    )
    embeddings = embedding_model.encode(result.column("text").to_pylist())
    embeddings_list = list(embeddings)  # Retain dtype
    result = result.append_column(
        "embedding",
        pa.array(embeddings_list, type=pa.list_(pa.from_numpy_dtype(embeddings.dtype))),
    )
    return result


def write_embeddings(
    target_dir: str,
    corpus_name: str,
    embeddings,  #: pa.Table,
    chunks_per_partition: int = 10000,
) -> pathlib.Path:
    """
    Write the embeddings produced by :func:`compute_embeddings()` to a directory of
    Parquet files on local disk.

    :param target_dir: Location where the files should be written (in a subdirectory)
    :param corpus_name: Corpus name to use in generating a location for writing files
    :param chunks_per_partition: Number of document chunks to write to each partition
      of the Parquet file

    :param returns: Path where data was written
    """
    # Third Party
    import pyarrow as pa
    import pyarrow.parquet as pq

    output_root = pathlib.Path(target_dir) / f"{corpus_name}.parquet"
    if os.path.exists(output_root):
        shutil.rmtree(output_root)
    os.makedirs(output_root)
    # Divide into bite-sized chunks
    for batch_num, batch in enumerate(embeddings.to_batches(chunks_per_partition)):
        pq.write_table(
            pa.Table.from_batches([batch]),
            output_root / f"part_{batch_num + 1:03d}.parquet",
        )
    return output_root


class InMemoryRetriever:
    """Simple retriever that keeps docs and embeddings in memory."""

    def __init__(
        self,
        data_file_or_table,  #: pathlib.Path | str | pa.Table,
        embedding_model_name: str,
    ):
        """
        :param data_file_or_table: Parquet file of document snippets and embeddings,
         or an equivalent in-memory PyArrow Table object.
         Should have columns `id`, `begin`, `end`, `text`, and `embedding`.
        :param embedding_model_name: Name of Sentence Transformers model to use for
         embeddings. Must be the same model that was used to compute embeddings in the
         data file.
        """
        # Third Party
        import pyarrow as pa
        import pyarrow.parquet as pq
        import sentence_transformers

        if isinstance(data_file_or_table, pa.Table):
            self._data_table = data_file_or_table
        else:
            self._data_table = pq.read_table(data_file_or_table)
        self._is_float16 = pa.types.is_float16(
            self._data_table.schema.field("embedding").type.value_type
        )
        self._embedding_model = sentence_transformers.SentenceTransformer(
            embedding_model_name,
            model_kwargs={"torch_dtype": "float16" if self._is_float16 else "float32"},
        )
        embeddings_array = np.array(
            list(self._data_table.column("embedding").to_numpy())
        )
        self._embeddings = torch.tensor(embeddings_array)

    def retrieve(self, query: str, top_k: int = 5) -> list[dict]:
        # Third Party
        import pyarrow as pa
        import sentence_transformers

        query_embeddings = self._embedding_model.encode(query)
        raw_result = sentence_transformers.util.semantic_search(
            query_embeddings, self._embeddings, top_k=top_k
        )

        # raw_result is a list of lists of {corpus_id, score}
        row_nums = [r["corpus_id"] for r in raw_result[0]]
        scores = [r["score"] for r in raw_result[0]]
        result = self._data_table.take(row_nums)
        result = result.append_column("score", pa.array(scores))
        return result.select(["id", "title", "url", "begin", "end", "text", "score"])


class RetrievalRequestProcessor(RequestProcessor):
    """
    Request processor that augments a chat completion request with relevant documents.
    """

    def __init__(self, retriever: Retriever, top_k: int = 3):
        """
        :param retriever: Interface to the vector database for retrieving document
            snippets.
        :param top_k: Number of snippets to retrieve
        """
        self._retriever = retriever
        self._top_k = top_k

    async def aprocess(
        self, inputs: ChatCompletionInputs
    ) -> list[ChatCompletionInputs]:
        last_turn_content = inputs.messages[-1].content
        retriever_output = self._retriever.retrieve(last_turn_content, self._top_k)

        # Retriever returns a PyArrow table. Currently we just use the texts and IDs of
        # the snippets.
        documents = [
            Document(doc_id=row["id"], text=row["text"])
            for row in retriever_output.to_pylist()
        ]
        return [inputs.model_copy(update={"documents": documents})]



# SPDX-License-Identifier: Apache-2.0
# ruff: noqa: E501

"""
Parser which receives Granite model output and returns the constituents of the output.

The input to the parser is assumed to be as follows:

    response_text

    # Citations:
    citations_text

    # Hallucinations:
    hallucinations_text

The output from the parser returns a dictionary as follows:

    "citations": List of citations
    "docs": List of document references
    "hallucinations": List of hallucinations
    "response": Model response text without the above constituents

"""

# Standard
import copy
import logging
import re
import sys

# Third Party
from nltk import sent_tokenize  # pylint: disable=import-error

# Local
from granite_io.io.consts import (
    _GRANITE_3_3_CITATIONS_START,
    _GRANITE_3_3_CITE_END,
    _GRANITE_3_3_CITE_START,
    _GRANITE_3_3_HALLUCINATIONS_START,
)

# Setup logger
logger = logging.getLogger("granite_io.io.granite_3_3.output_parser")
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(
    logging.Formatter("%(levelname)s %(asctime)s %(message)s", datefmt="%H:%M:%S")
)
logger.addHandler(handler)


def _find_substring_in_text(substring: str, text: str) -> list[dict[str, int]]:
    """
    Given two strings - substring and text - find and return all
    matches of substring within text. For each match return its begin and end index
    """
    span_matches = []

    matches_iter = re.finditer(re.escape(substring), text)
    for match in matches_iter:
        span_matches.append({"begin_idx": match.start(), "end_idx": match.end()})

    return span_matches


def _parse_hallucinations_text(hallucinations_text: str) -> list[dict]:
    """
    Given the hallucinations text output by model under the "# Hallucinations:"
    section, extract the hallucinations info as an array of the form:

    [
        {
            "hallucination_id": "Hallucination ID output by model",
            "risk": "Hallucination risk flag",
            "response_text": "Substring of response text for which
                                hallucination risk is computed"
        },
        ...
    ]
    """

    hallucinations = []

    # Find begin spans of all hallucinations
    matches_iter = re.finditer(
        "(\\d+)\\.\\s*Risk (low|high|unanswerable):",
        hallucinations_text,
    )
    matches = []
    for match in matches_iter:
        matches.append({"match_begin": match.start()})

    if len(matches) == 0:
        logger.warning(
            "Failed to extract hallucination info."
            "Expected hallucination info but none found."
        )

    # For each hallucination, extract its components (hallucination ID,
    # risk, response text)
    for i in range(len(matches)):  # pylint: disable=consider-using-enumerate
        cur_match = matches[i]

        # Select text corresponding to hallucination (which is the text from the
        # beginning of the hallucination until the beginning of the next
        # hallucination or the end of the text; whichever comes first)
        if i + 1 < len(matches):
            next_match_begin = matches[i + 1]["match_begin"] - 1
        else:
            next_match_begin = len(hallucinations_text)
        hallucination_str = hallucinations_text[
            cur_match["match_begin"] : next_match_begin
        ]

        # Within the hallucination text, extract the citation components
        # (hallucination ID, risk, response text)
        # Use ?s flag to include newlines in match
        matches_iter = re.finditer(
            "(?s)(\\d+)\\.\\s*Risk (low|high|unanswerable): (.+)$",
            hallucination_str,
        )
        idx = 0
        for match in matches_iter:
            cur_hallucination = {
                "hallucination_id": match.group(1),
                "risk": match.group(2),
                "response_text": match.group(3),
            }
            hallucinations.append(cur_hallucination)

            idx += 1

        if idx == 0:
            logger.warning(
                "Error in finding components of hallucination: "
                "Expected single RegEx match but found none."
            )
        if idx > 1:
            logger.warning(
                "Error in finding components of hallucination: "
                "Expected single RegEx match but found several."
            )

    return hallucinations


def _add_hallucination_response_spans(
    hallucination_info: list[dict], response_text_without_citations: str
) -> list[dict]:
    """
    Given the response text (cleaned from citation tags) and a
    parsed hallucinations info of the form:

    [
        {
            "hallucination_id": "Hallucination ID output by model",
            "risk": "Hallucination risk flag",
            "response_text": "Substring of response text for which hallucination
                                risk is computed"
        },
        ...
    ]

    add to each hallucination element in the array the following attributes
    (the "response_text" replaces the attribute of the same name):

        "response_text": "The response text corresponding to the hallucination
                            element cleaned from citation tags"
        "response_begin": "The begin index of "response_text" within the response
                            text (without citation tags)"
        "response_end": "The end index of "response_text" within the response
                            text (without citation tags)"
    """

    augmented_hallucination_info = copy.deepcopy(hallucination_info)

    for hallucination in augmented_hallucination_info:
        hallucination_response_text_without_citations = (
            _remove_citations_from_response_text(hallucination["response_text"])
        )
        matches = _find_substring_in_text(
            hallucination_response_text_without_citations,
            response_text_without_citations,
        )
        if len(matches) == 0:
            logger.warning(
                "Error in adding the response spans to hallucination: "
                "Hallucination text not found in response"
            )
            continue

        if len(matches) > 1:
            logger.warning(
                "Hallucination text found multiple times in "
                "response: Selecting first match"
            )
        hallucination["response_text"] = hallucination_response_text_without_citations
        hallucination["response_begin"] = matches[0]["begin_idx"]
        hallucination["response_end"] = matches[0]["end_idx"]

    return augmented_hallucination_info


def _parse_citations_text(citations_text: str) -> list[dict]:
    """
    Given the citations text output by model under the "# Citations:" section,
    extract the citation info as an array of the form:

    [
        {
            "citation_id": "Citation ID output by model",
            "doc_id": "ID of doc where the cited text is drawn from",
            "context_text": "The cited text from the context"
        },
        ...
    ]
    """

    citations = []

    # Find citations in the response
    pattern = r"(\d+):(.+)"
    matches_iter = re.finditer(pattern, citations_text)
    matches = []
    for match in matches_iter:
        matches.append({"match_begin": match.start()})

    if len(matches) == 0:
        logger.warning(
            "Error in extracting citation info. Expected citations but found none."
        )
        return citations

    # For each citation, extract its components (citation ID, doc ID, context text)
    for i in range(len(matches)):  # pylint: disable=consider-using-enumerate
        cur_match = matches[i]

        # Select text corresponding to citation (which is the text from the beginning
        # of the citation until the beginning of the next citation or the end of the
        # text; whichever comes first)
        if i + 1 < len(matches):
            next_match_begin = matches[i + 1]["match_begin"] - 1
        else:
            next_match_begin = len(citations_text)
        citation_str = citations_text[cur_match["match_begin"] : next_match_begin]

        # Within the citation text, extract the citation components
        # (citation ID, doc ID, context text)
        # Use ?s flag to include newlines in match
        pattern = r"(\d+):(.+)"
        matches_iter = re.finditer(
            pattern,
            citation_str,
        )
        idx = 0
        for match in matches_iter:
            context_text = match.group(2).strip().strip('"')

            cur_citation = {
                "citation_id": str(idx),
                "doc_id": match.group(1),
                "context_text": context_text,
            }
            citations.append(cur_citation)

            idx += 1

        if idx == 0:
            logger.warning(
                "Error in finding components of citation: Expected single RegEx match but found none."
            )
        if idx > 1:
            logger.warning(
                "Error in finding components of citation: Expected single RegEx match but found several."
            )

    return citations


def _add_citation_context_spans(
    citation_info: list[dict], docs: list[dict]
) -> list[dict]:
    """
    Given a set of docs and an array of citations of the form:

    [
        {
            "citation_id": "Citation ID output by model",
            "doc_id": "ID of doc where the cited text is drawn from",
            "context_text": "The cited text from the context"
        },
        ...
    ]

    add to each citation in the array the following two attributes:

        "context_begin": "The begin index of "context_text" within document with
                            ID doc_id"
        "context_end": "The end index of "context_text" within document with ID doc_id"
    """
    augmented_citation_info = copy.deepcopy(citation_info)
    docs_by_cit_doc_id = _create_dict(
        docs, citation_attrib="citation_id", document_attrib="doc_id"
    )
    for citation in augmented_citation_info:
        # Init values in event of error in processing
        citation["context_begin"] = 0
        citation["context_end"] = 0
        try:
            dict_id = str(citation["citation_id"]) + "-" + citation["doc_id"]
            doc = docs_by_cit_doc_id[dict_id]
        except KeyError:
            logger.warning(
                f"Document with id: {dict_id} not found "
                f"when adding citation context spans."
            )
            continue

        matches = _find_substring_in_text(citation["context_text"], doc["text"])
        if len(matches) == 0:
            logger.warning(
                "Error in adding the context spans to citation: "
                "Cited text not found in corresponding document"
            )
            continue

        if len(matches) > 1:
            logger.warning(
                "Cited text found multiple times in corresponding "
                "document: Selecting first match"
            )
        citation["context_begin"] = matches[0]["begin_idx"]
        citation["context_end"] = matches[0]["end_idx"]

    return augmented_citation_info


def _add_citation_response_spans(
    citation_info: list[dict],
    response_text_with_citations: str,
    response_text_without_citations: str,
) -> list[dict]:
    """
    Given the response text in two forms (the original with citation tags and
    the processed without) and an array of parsed citations of the form:

    [
        {
            "citation_id": "Citation ID output by model",
            "doc_id": "ID of doc where the cited text is drawn from",
            "context_text": "The cited text from the context",
            "context_begin": "The begin index of "context_text" within document with
                                ID doc_id" (Optional)
            "context_end": "The end index of "context_text" within document with
                                ID doc_id" (Optional)
        },
        ...
    ]

    add to each citation in the array the following two attributes:
        "response_text": "The substring of the response for which the citation is
                            provided"
        "response_begin": "The begin index of "response_text" within the response text"
        "response_end": "The end index of "response_text" within the response text"
    """

    augmented_citation_info = copy.deepcopy(citation_info)

    # Split response into sentences
    response_sentences = sent_tokenize(response_text_with_citations)

    # Create dictionary of the response sentence (cleaned from citations) corresponding
    # to each citation ID
    response_sents_by_citation_id = {}
    citation_idx = 0
    for sent_idx, sent in enumerate(response_sentences):
        # pylint: disable=anomalous-backslash-in-string
        pattern = f'{re.escape(_GRANITE_3_3_CITE_START)}{{"document_id": "(\d+)"}}{re.escape(_GRANITE_3_3_CITE_END)}'
        matches_iter = re.finditer(pattern, sent)
        for match in matches_iter:
            citation_id = str(citation_idx)
            if citation_idx > len(augmented_citation_info) - 1:
                augmented_citation_info.append({})  # Too many, add extra
            augmented_citation_info[citation_idx]["citation_id"] = citation_id
            augmented_citation_info[citation_idx]["document_id"] = match.group(1)
            citation_idx = citation_idx + 1
            if citation_id not in response_sents_by_citation_id:
                sent_without_citations = _remove_citations_from_response_text(sent)
                # Fixes for sentence splitting issue: Associate citation with previous
                # sentence if either of the following holds:
                # - The citation ID appears at the beginning of a sentence
                # - The found sentence is empty after removing citations
                if (match.start() == 0) or (len(sent_without_citations) == 0):
                    if sent_idx > 0:
                        sent_without_citations = _remove_citations_from_response_text(
                            response_sentences[sent_idx - 1]
                        )
                    else:
                        logger.warning(
                            "Error in extracting the response sentence of a citation: Found empty sentence"
                        )
                        response_sents_by_citation_id[citation_id] = ""
                        continue
                response_sents_by_citation_id[citation_id] = sent_without_citations
            else:
                logger.warning(
                    "Error in extracting the response sentence of a "
                    "citation: Citation ID appears in more than one "
                    "response sentences"
                )
                continue

    # For each citation bring the response sentence to which it refers and its
    # begin/end spans
    for i, citation in enumerate(augmented_citation_info):
        response_text = response_sents_by_citation_id.get(str(i), "")
        index = response_text_without_citations.find(response_text)
        if index < 0:
            logger.warning(
                "Error in extracting the response sentence of a citation: Unexpected error."
            )
            continue

        citation["response_text"] = response_text
        citation["response_begin"] = index
        citation["response_end"] = index + len(response_text_without_citations)

    return augmented_citation_info


def _get_docs_from_citations(docs: str) -> list[dict]:
    """
    Given a multi-line string with document information per line, extract
    and add to dictionary list with "doc_id" and "text" fields

    Document line format:
    1: "<text>"
    2: "<text>"
    <|something to ignore|>
    """
    doc_dicts = []
    if not docs or docs.isspace():
        return doc_dicts
    for i, line in enumerate(docs.splitlines()):
        if not line or line.isspace():
            continue
        line_split = line.split(":", maxsplit=1)
        if len(line_split) <= 1:
            logger.debug(f"Unable to retrieve doc text from: '{line}'.")
            continue
        doc_id = line_split[0].strip()
        if not doc_id.isdigit():
            logger.warning(f"Unable to retrieve doc id from: '{line}'.")
            continue
        text = line_split[1].strip().strip('"')

        # Using line index as citation_id
        doc_dicts.append({"citation_id": str(i), "doc_id": doc_id, "text": text})
    return doc_dicts


def _create_dict(input_array: object, **key_attrib_names: str) -> dict:
    """
    Given an array of dicts and the name of attribute(s) within the array, return a
    dict containing the contents of the array indexed by the given attribute(s)
    """
    new_dict = {}

    for item in input_array:
        new_dict_key_val: str = ""
        key_attribs_len = len(key_attrib_names)
        # Key for dictionary will be a combinations of attribute(s)
        # the dictionary that we are trying to index
        for key_attrib in key_attrib_names.values():
            new_dict_key_val += str(item[key_attrib])
            key_attribs_len -= 1
            if key_attribs_len > 0:
                new_dict_key_val += "-"

        if new_dict_key_val in new_dict:
            logger.warning(
                f"Found duplicate item while creating dictionary: "
                f"{new_dict[new_dict_key_val]}"
            )

        new_dict[new_dict_key_val] = item

    return new_dict


def _remove_citations_from_response_text(response_text: str) -> str:
    """
    Given a response text (potentially containing inline <co>...</co> tags),
    return the response text cleaned up from the <co>...</co> tags
    """

    # TODO:  Need a better way to turn on and then remove special tokens
    sop = "<|start_of_plugin|>"
    ret = response_text.replace(sop, "")
    eop = "<|end_of_plugin|>"
    ret = ret.replace(eop, "")
    pattern = f'{re.escape(_GRANITE_3_3_CITE_START)}{{"document_id": "\\d+"}}{re.escape(_GRANITE_3_3_CITE_END)}'
    return re.sub(pattern, "", ret).strip()


def _validate_response(response_text: str, citation_info: object):
    start = re.escape(_GRANITE_3_3_CITE_START)
    end = re.escape(_GRANITE_3_3_CITE_END)
    pattern = f"{start}(?:(?!({start}|{end})).)*{start}(?:(?!({start}|{end})).)*{end}"
    if re.search(pattern, response_text):
        logger.warning(f"Response contains nested citations: '{response_text}'.")

    opening_tag_count = response_text.count(_GRANITE_3_3_CITE_START)
    closing_tag_count = response_text.count(_GRANITE_3_3_CITE_END)

    if opening_tag_count != closing_tag_count:
        logger.warning(
            f"Response contains different number of cite start and end symbols: '{response_text}'."
        )

    if opening_tag_count != len(citation_info):
        logger.warning(
            f"Response contains different number of citations than those listed': {response_text}"
        )


def _split_model_output_into_parts(model_output: str) -> tuple[str, str, str]:
    """
    Divide the model output into its constituent parts: response_text, citations_text,
    and hallucinations_text.

    This assumes that the model output is of the following form:

    response_text
    # Citations
    citations_text
    # Hallucinations
    hallucinations_text

    where:
    - "TODO: # Citations" and "# Hallucinations> are literals, and
    - response_text, citations_text, hallucinations_text are variables corresponding
        the parts of the output

    Note:
    - The citations and hallucinations sections are both optional
    """
    citations_text = ""
    hallucinations_text = ""

    if (
        _GRANITE_3_3_HALLUCINATIONS_START in model_output
        and _GRANITE_3_3_CITATIONS_START not in model_output
    ):
        response_text, hallucinations_text = model_output.split(
            _GRANITE_3_3_HALLUCINATIONS_START
        )
    elif (
        _GRANITE_3_3_CITATIONS_START in model_output
        and _GRANITE_3_3_HALLUCINATIONS_START not in model_output
    ):
        response_text, citations_text = model_output.split(_GRANITE_3_3_CITATIONS_START)
    elif (
        _GRANITE_3_3_CITATIONS_START in model_output
        and _GRANITE_3_3_HALLUCINATIONS_START in model_output
    ):
        pre_citation_split, post_citation_split = model_output.split(
            _GRANITE_3_3_CITATIONS_START
        )
        if _GRANITE_3_3_HALLUCINATIONS_START in pre_citation_split:
            response_text, hallucinations_text = pre_citation_split.split(
                _GRANITE_3_3_HALLUCINATIONS_START
            )
            citations_text = post_citation_split
        else:
            citations_text, hallucinations_text = post_citation_split.split(
                _GRANITE_3_3_HALLUCINATIONS_START
            )
            response_text = pre_citation_split
    else:
        response_text = model_output

    # TODO:  Need a better way to turn on and then remove special tokens
    sop = "<|start_of_plugin|>"
    eop = "<|end_of_plugin|>"
    response_text = response_text.replace(sop, "").replace(eop, "")
    citations_text = citations_text.replace(sop, "").replace(eop, "")
    hallucinations_text = hallucinations_text.replace(sop, "").replace(eop, "")
    return response_text.strip(), citations_text.strip(), hallucinations_text.strip()


def _validate_spans_in_parser_output(parsed_task: object):
    """
    Validate that the hallucination and citation spans correspond to
    the model response
    """
    for hallucination in (
        parsed_task["hallucinations"] if parsed_task["hallucinations"] else []
    ):
        if "response_end" in hallucination and (
            hallucination["response_text"]
            != parsed_task["response"][
                hallucination["response_begin"] : hallucination["response_end"]
            ]
        ):
            logger.warning(
                "Hallucination span does not correspond to the model response."
            )
    for citation in parsed_task["citations"] if parsed_task["citations"] else []:
        docs_by_cit_doc_id = _create_dict(
            parsed_task["docs"], citation_attrib="citation_id", document_attrib="doc_id"
        )
        try:
            dict_id = citation["citation_id"] + "-" + citation["doc_id"]
            doc = docs_by_cit_doc_id[dict_id]
        except KeyError:
            logger.warning(
                f"Document with id: {dict_id} not found "
                f"when validation citation context spans."
            )
            continue
        if (
            citation["context_text"]
            != doc["text"][citation["context_begin"] : citation["context_end"]]
        ):
            logger.warning("Citation span does not correspond to the model response.")


def _update_docs_text_with_input_docs(
    docs_from_input: list[object], docs_from_citation: list[dict[str, str]]
) -> list[dict[str, str]]:
    """
    The documents passed in the chat completion call is the source of the documents
    used for the model output. The document text output by the model may not show
    the full context. Therefore, to have the full context, need to retrieve from the
    input passed to chat completion call, all documents and update the text in the
    citation documents to be aligned before finding the context spans.
    """

    augmented_docs_from_citation = copy.deepcopy(docs_from_citation)
    for citation_doc in augmented_docs_from_citation:
        for input_doc in docs_from_input:
            if citation_doc["text"].strip() in input_doc.text.strip():
                citation_doc["text"] = input_doc.text.strip()

    return augmented_docs_from_citation


def parse_model_output(
    model_output: str, docs_from_input: list[object]
) -> list[str | dict]:
    """
    Parse the constituents of the output (response) of a model into
    a format where they can be accessed individually

    Args:
        model_output: The response from model request
    Returns:
        Parsed part of the model output as follows:
            "docs": Document references
            "response": Model response without citations,
            "citations": Citations,
            "hallucinations": Hallucinations
    }
    """

    # Split model output into its parts: response, citation, and hallucination section
    response_text, citations_text, hallucinations_text = _split_model_output_into_parts(
        model_output
    )

    # Get documents from citations
    docs_from_citation = _get_docs_from_citations(citations_text)

    # Update 'docs_from_citation' with text from docs used as input to model prompt
    # as they are the full source of text. The full text is required to retrieve
    # context spans.
    docs = _update_docs_text_with_input_docs(docs_from_input, docs_from_citation)

    # Parsed response text
    response_text_without_citations = _remove_citations_from_response_text(
        response_text
    ).strip()

    # Parse hallucinations text
    if len(hallucinations_text) > 0:
        hallucination_info = _parse_hallucinations_text(hallucinations_text)
        augmented_hallucination_info = _add_hallucination_response_spans(
            hallucination_info, response_text_without_citations
        )
    else:
        augmented_hallucination_info = []
    logger.debug(f"Parsed hallucination info:\n{augmented_hallucination_info}\n")

    # Parse citations text
    if len(citations_text) > 0:
        citation_info = _parse_citations_text(citations_text)
        citation_info_with_context_spans = _add_citation_context_spans(
            citation_info, docs
        )
        citation_info_with_context_response_spans = _add_citation_response_spans(
            citation_info_with_context_spans,
            response_text,
            response_text_without_citations,
        )
        _validate_response(response_text, citation_info)
    else:
        citation_info_with_context_response_spans = []
    logger.debug(
        f"Parsed citation info:\n{citation_info_with_context_response_spans}\n"
    )

    # Join all objects into single output
    result = {
        "docs": docs if docs else None,
        "response": response_text_without_citations,
        "citations": (
            citation_info_with_context_response_spans
            if citation_info_with_context_response_spans
            else None
        ),
        "hallucinations": (
            augmented_hallucination_info if augmented_hallucination_info else None
        ),
    }
    logger.debug(f"Combined parser output:\n{result}\n")

    # Validate spans in parser output by checking if the citation/response text
    # matches the begin/end spans
    _validate_spans_in_parser_output(result)

    return result



# SPDX-License-Identifier: Apache-2.0

# Standard
import datetime
import json

# Third Party
from pydantic_core import PydanticCustomError
import pydantic

# Local
from granite_io.io.base import InputProcessor
from granite_io.io.consts import (
    _GRANITE_3_2_2B_HF,
    _GRANITE_3_2_2B_OLLAMA,
    _GRANITE_3_2_COT_END,
    _GRANITE_3_2_COT_START,
    _GRANITE_3_2_MODEL_NAME,
)
from granite_io.io.registry import input_processor
from granite_io.types import (
    AssistantMessage,
    ChatCompletionInputs,
    SystemMessage,
    ToolResultMessage,
    UserMessage,
)

_TODAYS_DATE_STR = datetime.datetime.now().strftime("%B %d, %Y")


def override_date_for_testing(todays_date_str: str | None):
    """Override the date that methods in this file will use for today's date, in order
    to make test outputs consistent.

    :param todays_date_str: Date string to use for generating prompts until further
     notice, or ``None`` to revert to using the real date.
    """
    global _TODAYS_DATE_STR  # pylint: disable=global-statement
    if todays_date_str is None:
        _TODAYS_DATE_STR = datetime.datetime.now().strftime("%B %d, %Y")
    else:
        _TODAYS_DATE_STR = todays_date_str


# String that comes at the beginning of the system message that a Granite 3.2 model must
# receive at the beginning of the prompt for any completion request that does not
# provide a custom system message.
#
# Note that the original Jinja template tends to choose weird dates from the future for
# the "Today's date" part. Instead of replicating that behavior, we put today's actual
# date in that section of the prompt. This difference probably doesn't matter, since
# none of the supervised fine tuning data exercises knowledge cutoffs.
#
# As an additional wrinkle, we need to use a consistent date when testing, so we use a
# function to recreate this string every time we need it.
def _make_system_message_start():
    return f"""\
Knowledge Cutoff Date: April 2024.
Today's Date: {_TODAYS_DATE_STR}.
You are Granite, developed by IBM."""


# String that a Granite 3.2 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are both tools and RAG documents in the current request.
_TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART = """\
 You are a helpful AI assistant with access to the following tools. When a tool is \
required to answer the user's query, respond with <|tool_call|> followed by a JSON \
list of tools used. If a tool does not exist in the provided list of tools, notify the \
user that you do not have the ability to fulfill the request.

Write the response to the user's input by strictly aligning with the facts in the \
provided documents. If the information needed to answer the question is not available \
in the documents, inform the user that the question cannot be answered based on the \
available data."""

# String that a Granite 3.2 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are documents in the current request but there are no tools in the current
# request.
_NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART = """\
Write the response to the user's input by strictly aligning with the facts in the \
provided documents. If the information needed to answer the question is not available \
in the documents, inform the user that the question cannot be answered based on the \
available data."""

# String that a Granite 3.2 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are tools in the current request but there are no documents in the current
# request.
_TOOLS_AND_NO_DOCS_SYSTEM_MESSAGE_PART = """\
You are a helpful AI assistant with access to the following tools. When a tool is \
required to answer the user's query, respond with <|tool_call|> followed by a JSON \
list of tools used. If a tool does not exist in the provided list of tools, notify the \
user that you do not have the ability to fulfill the request."""

# String that a Granite 3.2 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are no tools or documents in the current request and the "thinking" flag is
# set to `True`.
_NO_TOOLS_AND_NO_DOCS_AND_THINKING_SYSTEM_MESSAGE_PART = f"""\
 You are a helpful AI assistant.
Respond to every user query in a comprehensive and detailed way. You can write down \
your thoughts and reasoning process before responding. In the thought process, engage \
in a comprehensive cycle of analysis, summarization, exploration, reassessment, \
reflection, backtracing, and iteration to develop well-considered thinking process. \
In the response section, based on various attempts, explorations, and reflections from \
the thoughts section, systematically present the final solution that you deem correct. \
The response should summarize the thought process. Write your thoughts after '\
{_GRANITE_3_2_COT_START}' and write your response after '{_GRANITE_3_2_COT_END}' \
for each user query."""

# String that a Granite 3.2 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are no tools or documents in the current request and the "thinking" flag is
# set to `False`.
_NO_TOOLS_NO_DOCS_NO_THINKING_SYSTEM_MESSAGE_PART = """\
 You are a helpful AI assistant."""


# String that a Granite 3.2 model must receive immediately after either
# _TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE  (if there are tools) or
# _NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE (if there are no tools) in the system prompt
# if the "citations" flag is `True` and there are documents.
_DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART = """\


In your response, use the symbols <co> and </co> to indicate when a fact comes from a \
document in the search result, e.g <co>0</co> for a fact from document 0. Afterwards, \
list all the citations with their corresponding documents in an ordered list."""

# String that a Granite 3.2 model must receive immediately after either
# _TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE (if there are tools and no citations) or
# _NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE (if there are no tools or citations) or
# _DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART in the system prompt
# if the "hallucinations" flag is `True` and there are documents.
# Note that a list of zero documents counts as "having documents".
_DOCS_AND_HALLUCINATIONS_SYSTEM_MESSAGE_PART = """\


Finally, after the response is written, include a numbered list of sentences from the \
response that are potentially hallucinated and not based in the documents."""


class Document(pydantic.BaseModel):
    text: str


class ControlsRecord(pydantic.BaseModel):
    citations: bool | None = None
    hallucinations: bool | None = None
    length: str | None = None  # Length output control variable
    originality: str | None = None

    @pydantic.field_validator("length", mode="after")
    @classmethod
    def _validate_length(cls, value: str | None) -> str | None:
        if value is None or value == "short" or value == "long":
            return value
        raise PydanticCustomError(
            "length field validator",
            'length ({length}) must be "short" or "long" or None',
            {"length": value},
        )

    @pydantic.field_validator("originality", mode="after")
    @classmethod
    def _validate_originality(cls, value: str | None) -> str | None:
        if value is None or value == "extractive" or value == "abstractive":
            return value
        raise PydanticCustomError(
            "originality field validator",
            'originality ({originality}) must be "extractive" or "abstractive" or None',
            {"originality": value},
        )


class Granite3Point2Inputs(ChatCompletionInputs):
    """
    Class that represents the inputs to a Granite 3.2 model generation call.

    Contains fields for all input functionality supported by the current version of
    Granite.

    This class will gain additional fields as new functionality is added to Granite.
    """

    documents: list[Document] = []
    controls: ControlsRecord | None = None

    thinking: bool = False

    @pydantic.field_validator("messages")
    @classmethod
    def _validate_inputs_messages(cls, messages: list) -> list:
        # Make a copy so the validation code below can mutate the messages list but pass
        # through the original value. The caller also might have a pointer to the list.
        original_messages = messages
        messages = messages.copy()

        # There is no supervised fine tuning data for the case of zero messages.
        # Models are not guaranteed to produce a valid response if there are zero
        # messages.
        if len(messages) == 0:
            raise ValueError(
                "No messages. Model behavior for this case is not defined."
            )

        # The first message, and only the first message, may be the system message.
        first_message_is_system_message = isinstance(messages[0], SystemMessage)
        if first_message_is_system_message:
            messages = messages[1:]
            # If there is a system message, there must be at least one more user or
            # assistant message.
            if len(messages) == 0:
                raise ValueError(
                    "Input contains only a system message. Model behavior for this "
                    "case is not defined."
                )

        # The first message that is not a system message must be
        # either a user or assistant message.
        if not isinstance(messages[0], UserMessage | AssistantMessage):
            if first_message_is_system_message:
                raise ValueError(
                    f"First message after system message must be a user or "
                    f"assistant message. Found type {type(messages[0])}"
                )
            raise ValueError(
                f"First message must be a system, user, or assistant "
                f"Found type {type(messages[0])}"
            )

        # Undocumented constraint: All other messages form a conversation that
        # alternates strictly between user and assistant, possibly with tool calls
        # after an assistant turn and before the next user turn.
        # TODO: Validate this invariant.

        # Pydantic will use the value that this validator returns as the value of the
        # messages field. Undo any changes that we made during validation and return
        # the original value.
        return original_messages


@input_processor(
    _GRANITE_3_2_MODEL_NAME,
    # Huggingface
    _GRANITE_3_2_2B_HF,
    "ibm-granite/granite-3.2-8b-instruct",
    # Ollama
    "granite3.2",
    "granite3.2:8b",
    _GRANITE_3_2_2B_OLLAMA,
    # RITS
    "ibm-granite/granite-8b-instruct-preview-4k",
)
class Granite3Point2InputProcessor(InputProcessor):
    """
    Input processor for version 3.2 of the main Granite models, all sizes.

    This input processor is based on the Jinja template that was used during
    supervised fine tuning of these models. This template is as follows:
    ```
    {%- if messages[0]['role'] == 'system' %}
        {%- set system_message = messages[0]['content'] %}
        {%- set loop_messages = messages[1:] %}
    {%- else %}
        {%- set system_message = \"Knowledge Cutoff Date: April 2024.\nToday's Date: \"
          + strftime_now('%B %d, %Y') + \".\nYou are Granite, developed by IBM.\" %}
        {%- if tools and documents %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant with access to the following tools.
                  When a tool is required to answer the user's query, respond with
                  <|tool_call|> followed by a JSON list of tools used. If a tool does
                  not exist in the provided list of tools, notify the user that you do
                  not have the ability to fulfill the request.\n\nWrite the response to
                  the user's input by strictly aligning with the facts in the provided
                  documents. If the information needed to answer the question is not
                  available in the documents, inform the user that the question cannot
                  be answered based on the available data.\" %}
        {%- elif tools %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant with access to the following tools. When a tool is required to
                answer the user's query, respond with <|tool_call|> followed by a JSON
                list of tools used. If a tool does not exist in the provided list of
                tools, notify the user that you do not have the ability to fulfill the
                request.\" %}
        {%- elif documents %}
                {%- set system_message = system_message + \" Write the response to the
                user's input by strictly aligning with the facts in the provided
                documents. If the information needed to answer the question is not
                available in the documents, inform the user that the question cannot be
                answered based on the available data.\" %}
        {%- elif thinking %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant.\nRespond to every user query in a comprehensive and detailed
                way. You can write down your thoughts and reasoning process before
                responding. In the thought process, engage in a comprehensive cycle of
                analysis, summarization, exploration, reassessment, reflection,
                backtracing, and iteration to develop well-considered thinking process.
                In the response section, based on various attempts, explorations, and
                reflections from the thoughts section, systematically present the final
                solution that you deem correct. The response should summarize the
                thought process. Write your thoughts after 'Here is my thought process:'
                and write your response after 'Here is my response:' for each user
                query.\" %}
        {%- else %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant.\" %}
        {%- endif %}
        {%- if 'citations' in controls and documents %}
            {%- set system_message = system_message + '\n\nIn your response, use the
            symbols <co> and </co> to indicate when a fact comes from a document in the
            search result, e.g <co>0</co> for a fact from document 0. Afterwards, list
            all the citations with their corresponding documents in an ordered list.' %}
        {%- endif %}
        {%- if 'hallucinations' in controls and documents %}
            {%- set system_message = system_message + '\n\nFinally, after the response
            is written, include a numbered list of sentences from the response that are
            potentially hallucinated and not based in the documents.' %}
        {%- endif %}
        {%- set loop_messages = messages %}
    {%- endif %}
    {{- '<|start_of_role|>system<|end_of_role|>' + system_message +
        '<|end_of_text|>\n' }}
    {%- if tools %}
        {{- '<|start_of_role|>tools<|end_of_role|>' }}
        {{- tools | tojson(indent=4) }}
        {{- '<|end_of_text|>\n' }}
    {%- endif %}
    {%- if documents %}
        {{- '<|start_of_role|>documents<|end_of_role|>' }}
        {%- for document in documents %}
            {{- 'Document ' + loop.index0 | string + '\n' }}
            {{- document['text'] }}
            {%- if not loop.last %}
                {{- '\n\n'}}
            {%- endif%}
        {%- endfor %}
        {{- '<|end_of_text|>\n' }}
    {%- endif %}
    {%- for message in loop_messages %}
        {{- '<|start_of_role|>' + message['role'] + '<|end_of_role|>' +
        message['content'] + '<|end_of_text|>\n' }}
        {%- if loop.last and add_generation_prompt %}
            {{- '<|start_of_role|>assistant' }}
            {%- if controls %}
                {{- ' ' + controls | tojson()}}
            {%- endif %}
            {{- '<|end_of_role|>' }}
        {%- endif %}
    {%- endfor %}
    ```
    """

    def _split_messages(
        self, inputs: Granite3Point2Inputs
    ) -> tuple[SystemMessage | None, list[UserMessage]]:
        """
        Separate the system message from other messages.

        :returns: Tuple of system message, if present, and remaining messages.
        """
        messages = inputs.messages

        # Validation code in the Inputs class should already have verified that there
        # are either zero or one system messages, and that the system message, if
        # present, occurs at position zero.
        if isinstance(messages[0], SystemMessage):
            # First message is a system message.
            return messages[0], messages[1:]
        return None, messages

    def _build_default_system_message(self, inputs: Granite3Point2Inputs) -> str:
        """
        :param inputs: All inputs to a completion request that does not include a custom
            system message.
        :returns: The standard system message portion of the prompt for the request,
            as a string suitable to feed to the model's tokenizer.
        """
        # Compute the predicates that determine exactly what default system message to
        # use.
        have_documents = inputs.documents is not None and len(inputs.documents) > 0
        have_tools = len(inputs.tools) > 0

        # Carefully hew to the policy that the original Jinja template's behavior
        # defines.
        # First, disallow the cases that the authors of the Jinja template did not
        # provide any code to handle.
        if inputs.thinking and have_documents:
            raise ValueError(
                f"'thinking' flag is set, but documents were provided. "
                f"{_GRANITE_3_2_MODEL_NAME} only supports the 'thinking' flag when "
                f"documents are not provided."
            )
        if inputs.thinking and have_tools:
            raise ValueError(
                f"'thinking' flag is set, but tools were provided. "
                f"{_GRANITE_3_2_MODEL_NAME} only supports the 'thinking' flag when "
                f"tools are not provided."
            )

        # The default system message starts with a header that includes the date and
        # knowledge cutoff.
        system_message = "<|start_of_role|>system<|end_of_role|>"
        system_message += _make_system_message_start()

        # Add a middle part that varies depending on tools, documents, and citations.
        if have_documents and have_tools:
            system_message += _TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART
        elif have_documents:  # and not have_tools
            system_message += _NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART
        elif have_tools:  # and not have_documents
            system_message += _TOOLS_AND_NO_DOCS_SYSTEM_MESSAGE_PART
        elif inputs.thinking:  # if not have_documents and not have_tools
            system_message += _NO_TOOLS_AND_NO_DOCS_AND_THINKING_SYSTEM_MESSAGE_PART
        else:  # if not inputs.thinking and not have_documents and not have_tools
            system_message += _NO_TOOLS_NO_DOCS_NO_THINKING_SYSTEM_MESSAGE_PART

        # Next comes an optional section of instructions for citations.
        if inputs.controls and inputs.controls.citations:
            if not have_documents:
                # TODO: The template skips the citations instruction in this case.
                # Is this behavior an error? Should we raise an error if the caller
                # sets the citations flag but provides zero documents?
                pass
            else:  # if have_documents
                system_message += _DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART

        # Then comes an optional section of instructions for hallucinations.
        if inputs.controls and inputs.controls.hallucinations:
            if not have_documents:
                raise ValueError(
                    f"'hallucinations' flag is set, but the model input does not "
                    f"include documents. {_GRANITE_3_2_MODEL_NAME} only supports the "
                    f"'hallucinations' flag when documents are provided."
                )
            # if have_documents
            system_message += _DOCS_AND_HALLUCINATIONS_SYSTEM_MESSAGE_PART

        # Finish with an end of text
        system_message += "<|end_of_text|>\n"

        return system_message

    def _message_to_prompt_string(self, message: UserMessage | AssistantMessage) -> str:
        if isinstance(message, UserMessage):
            return (
                f"<|start_of_role|>user<|end_of_role|>{message.content}"
                f"<|end_of_text|>\n"
            )
        if isinstance(message, AssistantMessage):
            # Note that we discard any tool calls in the message, per the Jinja
            # template.
            return (
                f"<|start_of_role|>assistant<|end_of_role|>{message.content}"
                f"<|end_of_text|>\n"
            )
        if isinstance(message, ToolResultMessage):
            # Note that we discard the tool call ID, per the Jinja template.
            return (
                f"<|start_of_role|>tool<|end_of_role|>{message.content}"
                f"<|end_of_text|>\n"
            )
        raise TypeError(f"Unexpected message type {type(message)}")

    def _build_controls_record(self, inputs: Granite3Point2Inputs) -> dict | None:
        """
        Use the output control flags in ``inputs`` to build a version of the
        undocumented arbitrary JSON data regarding output controls that the Jinja
        template expected to see in the input for each chat completion request.

        :returns: A fake JSON record for "controls", or nothing of no output control
        flags were set.
        """
        if not inputs.controls:
            return None
        result = {}
        if inputs.controls.citations:
            # The following is a guess; we have no example data for this case.
            result["citations"] = True
        if inputs.controls.hallucinations:
            # The following is a guess; we have no example data for this case.
            result["hallucinations"] = True
        if inputs.controls.length is not None:
            result["length"] = inputs.controls.length
        if inputs.controls.originality is not None:
            result["originality"] = inputs.controls.originality

        if len(result) == 0:
            return None
        return result

    def transform(
        self, inputs: ChatCompletionInputs, add_generation_prompt: bool = True
    ) -> str:
        # Downcast to a Granite-specific request type with possible additional fields.
        # This operation also performs additional validation.
        inputs = Granite3Point2Inputs.model_validate(inputs.model_dump())

        # Check for a caller-provided system message
        system_message_json, loop_messages = self._split_messages(inputs)

        if system_message_json is not None:
            if inputs.thinking:
                raise ValueError(
                    f"'thinking' flag is set, but the model input includes a custom "
                    f"system message. {_GRANITE_3_2_MODEL_NAME} only supports the "
                    f"'thinking' flag when the default system message is used."
                )
            if len(inputs.documents) > 0:
                raise ValueError(
                    f"The model input includes documents and a custom system message. "
                    f"{_GRANITE_3_2_MODEL_NAME} only supports the documents list when "
                    f"the default system message is used."
                )
            if inputs.controls and inputs.controls.citations:
                raise ValueError(
                    f"'citations' flag is set, but the model input includes a custom "
                    f"system message. {_GRANITE_3_2_MODEL_NAME} only supports the "
                    f"'citations' flag when the default system message is used."
                )
            if inputs.controls and inputs.controls.hallucinations:
                raise ValueError(
                    f"'hallucinations' flag is set, but the model input includes a "
                    f"custom system message. {_GRANITE_3_2_MODEL_NAME} only supports "
                    f"the 'hallucinations' flag when the default system message is "
                    f"used."
                )
            system_message = (
                f"<|start_of_role|>system<|end_of_role|>"
                f"{system_message_json.content}<|end_of_text|>\n"
            )
        else:  # if system_message_json is None:
            # No caller-provided system message.
            # Create a default system message according to the rules implied by the
            # tokenizer's Jinja template.
            system_message = self._build_default_system_message(inputs)

        if len(inputs.tools) == 0:
            tools_part = ""
        else:
            tools_part = (
                "<|start_of_role|>tools<|end_of_role|>"
                + json.dumps(inputs.tools.to_openai_json(), indent=4)
                + "<|end_of_text|>\n"
            )

        if len(inputs.documents) == 0:
            documents_part = ""
        else:
            documents_body = "\n\n".join(
                [
                    f"Document {i}\n{inputs.documents[i].text}"
                    for i in range(len(inputs.documents))
                ]
            )
            documents_part = (
                "<|start_of_role|>documents<|end_of_role|>"
                + documents_body
                + "<|end_of_text|>\n"
            )

        messages_part = "".join(
            [self._message_to_prompt_string(message) for message in loop_messages]
        )

        # Jinja template expects arbitrary JSON, while our dataclass has specific
        # fields for supported controls.
        controls_record = self._build_controls_record(inputs)
        controls_str = (
            "" if controls_record is None else " " + json.dumps(controls_record)
        )

        generation_prompt_part = (
            ""
            if not add_generation_prompt
            else f"<|start_of_role|>assistant{controls_str}<|end_of_role|>"
        )

        return (
            system_message
            + tools_part
            + documents_part
            + messages_part
            + generation_prompt_part
        )



# SPDX-License-Identifier: Apache-2.0

"""
This example shows how to infer or call a model using the framework
and an Ollama backend to serve the model.
In this scenario, it is a simple RAG use case where citations and
hallucinations are enabled.
"""

# Standard
import pprint

# Local
from granite_io import make_backend, make_io_processor
from granite_io.types import ChatCompletionInputs, UserMessage

model_name = "granite3.2:8b"
io_processor = make_io_processor(
    model_name, backend=make_backend("openai", {"model_name": model_name})
)
question = "What is RAG?"
messages = [UserMessage(content=question)]

documents = [
    {
        "text": "Dynamic Data: Automatically generate and refine graphs to best fit your domain and ontology needs."  # noqa: E501
    },
    {
        "text": "RAG, retrieval-augmented generation, is a technique that grants generative artificial intelligence models information retrieval capabilities."  # noqa: E501
    },
]

# With RAG and citations
outputs = io_processor.create_chat_completion(
    ChatCompletionInputs(
        messages=messages,
        documents=documents,
        controls={"citations": True, "hallucinations": True},
    )
)
print("\n\n>> Model raw output:\n")
response = outputs.results[0].next_message.raw
print(response)

print("\n\n>> Response:\n")
response = outputs.results[0].next_message.content
print(response)

if outputs.results[0].next_message.citations:
    print("\n\n>> Citations:\n")
    pprint.pprint(outputs.results[0].next_message.citations, sort_dicts=False)

    print("\n\n>> Documents:\n")
    pprint.pprint(outputs.results[0].next_message.documents, sort_dicts=False)

if outputs.results[0].next_message.hallucinations:
    print("\n\n>> Hallucinations:\n")
    pprint.pprint(outputs.results[0].next_message.hallucinations, sort_dicts=False)



# SPDX-License-Identifier: Apache-2.0
# ruff: noqa: E501

"""
Tests for the model output parser
"""

# Standard
import os

# Third Party
from test_utils import load_text_file

# Local
from granite_io.io.granite_3_2.input_processors.granite_3_2_input_processor import (
    Document,
)
from granite_io.io.granite_3_2.output_processors.granite_3_2_output_parser import (
    parse_model_output,
)

TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "testdata")


def test_output():
    model_output = load_text_file(os.path.join(TEST_DATA_DIR, "test_output.txt"))
    parsed_output = parse_model_output(model_output, "")

    assert parsed_output
    assert isinstance(parsed_output, dict)
    keys = ["docs", "response", "citations", "hallucinations"]
    for key in keys:
        assert key in parsed_output

    response = parsed_output["response"]
    assert isinstance(response, str)
    assert "To efficiently find the fastest way for a seller" in response

    assert parsed_output["docs"] is None
    assert parsed_output["citations"] is None
    assert parsed_output["hallucinations"] is None


def test_output_with_citation():
    model_output = load_text_file(
        os.path.join(TEST_DATA_DIR, "test_output_with_citation.txt")
    )
    parsed_output = parse_model_output(model_output, "")

    assert parsed_output
    assert isinstance(parsed_output, dict)
    keys = ["docs", "response", "citations", "hallucinations"]
    for key in keys:
        assert key in parsed_output

    docs = parsed_output["docs"]
    assert len(docs) == 1
    doc = docs[0]  # pylint: disable = unsubscriptable-object
    keys = ["doc_id", "text"]
    for key in keys:
        assert key in doc
    assert doc["doc_id"] == "1"
    assert "RAG, retrieval-augmented generation" in doc["text"]

    response = parsed_output["response"]
    assert isinstance(response, str)
    assert "It's a technique that combines the strengths" in response

    citations = parsed_output["citations"]
    assert len(citations) == 1
    citation = citations[0]
    keys = [
        "citation_id",
        "doc_id",
        "context_text",
        "context_begin",
        "context_end",
        "response_text",
        "response_begin",
        "response_end",
    ]
    for key in keys:
        assert key in citation
    assert citation["citation_id"] == "1"
    assert citation["doc_id"] == "1"
    assert len(citation["context_text"]) >= 1
    assert citation["context_begin"] == 0
    assert citation["context_end"] == 141
    assert len(citation["response_text"]) >= 1
    assert citation["response_begin"] == 0
    assert citation["response_end"] == 46

    assert parsed_output["hallucinations"] is None


def test_output_with_invalid_citation():
    model_output = load_text_file(
        os.path.join(TEST_DATA_DIR, "test_output_with_invalid_citation.txt")
    )
    parsed_output = parse_model_output(model_output, "")

    assert parsed_output["citations"] is None
    assert parsed_output["hallucinations"] is None


def test_output_with_colons_in_citation_text():
    model_output = load_text_file(
        os.path.join(TEST_DATA_DIR, "test_output_with_colons_citation_text.txt")
    )
    parsed_output = parse_model_output(model_output, "")

    assert parsed_output
    assert isinstance(parsed_output, dict)
    keys = ["docs", "response", "citations", "hallucinations"]
    for key in keys:
        assert key in parsed_output

    docs = parsed_output["docs"]
    assert len(docs) == 1
    doc = docs[0]  # pylint: disable = unsubscriptable-object
    keys = ["doc_id", "text"]
    for key in keys:
        assert key in doc
    assert doc["doc_id"] == "2"
    assert (
        "3부 : [UNK] [UNK] 종료 [UNK], 공영 주택의 세입자는 이 프로그램에서 [UNK] [UNK] 종료에 직면합니다. 이 2가지 상황에 대해 더 자세"
        in doc["text"]
    )

    response = parsed_output["response"]
    assert isinstance(response, str)
    assert "Yes, you can visit a law library to conduct your legal research" in response

    citations = parsed_output["citations"]
    assert len(citations) == 1
    citation = citations[0]
    keys = [
        "citation_id",
        "doc_id",
        "context_text",
        "context_begin",
        "context_end",
        "response_text",
        "response_begin",
        "response_end",
    ]
    for key in keys:
        assert key in citation
    assert citation["citation_id"] == "1"
    assert citation["doc_id"] == "2"
    assert len(citation["context_text"]) >= 1

    hallucinations = parsed_output["hallucinations"]
    assert len(hallucinations) == 4
    hallc_id = 1
    for hallucination in hallucinations:
        keys = [
            "hallucination_id",
            "risk",
            "response_text",
            "response_begin",
            "response_end",
        ]
        for key in keys:
            assert key in hallucination
        assert hallucination["hallucination_id"] == str(hallc_id)
        assert len(hallucination["response_text"]) >= 1
        hallc_id += 1


def test_output_with_citation_hallucinations():
    model_output = load_text_file(
        os.path.join(TEST_DATA_DIR, "test_output_with_citation_hallucinations.txt")
    )
    parsed_output = parse_model_output(model_output, "")

    assert parsed_output
    assert isinstance(parsed_output, dict)
    keys = ["docs", "response", "citations", "hallucinations"]
    for key in keys:
        assert key in parsed_output

    docs = parsed_output["docs"]
    assert len(docs) == 1
    doc = docs[0]  # pylint: disable = unsubscriptable-object
    keys = ["doc_id", "text"]
    for key in keys:
        assert key in doc
    assert doc["doc_id"] == "1"
    assert "RAG, retrieval-augmented generation" in doc["text"]

    response = parsed_output["response"]
    assert isinstance(response, str)
    assert len(response) >= 1

    citations = parsed_output["citations"]
    assert len(citations) == 1
    citation = citations[0]
    keys = [
        "citation_id",
        "doc_id",
        "context_text",
        "context_begin",
        "context_end",
        "response_text",
        "response_begin",
        "response_end",
    ]
    for key in keys:
        assert key in citation
    assert citation["citation_id"] == "1"
    assert citation["doc_id"] == "1"
    assert len(citation["context_text"]) >= 1
    assert citation["context_begin"] == 0
    assert citation["context_end"] == 141
    assert len(citation["response_text"]) >= 1
    assert citation["response_begin"] == 47
    assert citation["response_end"] == 194

    hallucinations = parsed_output["hallucinations"]
    assert len(hallucinations) == 3
    hallc_id = 1
    for hallucination in hallucinations:
        keys = [
            "hallucination_id",
            "risk",
            "response_text",
            "response_begin",
            "response_end",
        ]
        for key in keys:
            assert key in hallucination
        assert hallucination["hallucination_id"] == str(hallc_id)
        assert hallucination["risk"] == "low"
        assert len(hallucination["response_text"]) >= 1
        hallc_id += 1


def test_output_with_citation_from_source():
    model_output = load_text_file(
        os.path.join(TEST_DATA_DIR, "test_output_with_citation_from_source.txt")
    )
    doc_source = load_text_file(os.path.join(TEST_DATA_DIR, "test_document_source.txt"))
    doc_input = [Document(text=f"{doc_source}")]
    parsed_output = parse_model_output(model_output, doc_input)

    assert parsed_output
    assert isinstance(parsed_output, dict)
    keys = ["docs", "response", "citations", "hallucinations"]
    for key in keys:
        assert key in parsed_output

    docs = parsed_output["docs"]
    assert len(docs) == 1
    doc = docs[0]  # pylint: disable = unsubscriptable-object
    keys = ["doc_id", "text"]
    for key in keys:
        assert key in doc
    assert doc["doc_id"] == "0"
    assert doc["text"] == doc_input[0].text

    response = parsed_output["response"]
    assert isinstance(response, str)
    assert "This consistent achievement makes Limerick stand out" in response

    citations = parsed_output["citations"]
    assert len(citations) == 1
    citation = citations[0]
    keys = [
        "citation_id",
        "doc_id",
        "context_text",
        "context_begin",
        "context_end",
        "response_text",
        "response_begin",
        "response_end",
    ]
    for key in keys:
        assert key in citation
    assert citation["citation_id"] == "1"
    assert citation["doc_id"] == "0"
    assert len(citation["context_text"]) >= 1
    assert citation["context_begin"] == 0
    assert citation["context_end"] == 561
    assert len(citation["response_text"]) >= 1
    assert citation["response_begin"] == 300
    assert citation["response_end"] == 379

    assert parsed_output["hallucinations"] is None


def test_output_with_multiple_citations_per_document():
    model_output = load_text_file(
        os.path.join(
            TEST_DATA_DIR, "test_output_with_multiple_citations_per_document.txt"
        )
    )
    parsed_output = parse_model_output(model_output, "")

    assert parsed_output
    assert isinstance(parsed_output, dict)
    keys = ["docs", "response", "citations", "hallucinations"]
    for key in keys:
        assert key in parsed_output

    docs = parsed_output["docs"]
    assert len(docs) == 2
    doc = docs[0]  # pylint: disable = unsubscriptable-object
    keys = ["doc_id", "text"]
    for key in keys:
        assert key in doc
    assert doc["doc_id"] == "2"
    assert doc["citation_id"] == "1"
    assert "law - libraries / 에서 [UNK] 수 [UNK] 지역 법률 도서관에" in doc["text"]
    doc = docs[1]  # pylint: disable = unsubscriptable-object
    keys = ["doc_id", "text"]
    for key in keys:
        assert key in doc
    assert doc["doc_id"] == "2"
    assert doc["citation_id"] == "2"
    assert (
        "1부 : 머리말 이 시리즈의 비디오에 대한 나머지 부분을 [UNK] 머리말을 들으려면"
        in doc["text"]
    )

    response = parsed_output["response"]
    assert isinstance(response, str)
    assert (
        "Law libraries typically have self-help legal books, official reports, and case laws that could be helpful for your research"
        in response
    )

    citations = parsed_output["citations"]
    assert len(citations) == 2
    citation = citations[0]
    keys = [
        "citation_id",
        "doc_id",
        "context_text",
        "context_begin",
        "context_end",
        "response_text",
        "response_begin",
        "response_end",
    ]
    for key in keys:
        assert key in citation
    assert citation["citation_id"] == "1"
    assert citation["doc_id"] == "2"
    assert len(citation["context_text"]) >= 1

    hallucinations = parsed_output["hallucinations"]
    assert len(hallucinations) == 4
    hallc_id = 1
    for hallucination in hallucinations:
        keys = [
            "hallucination_id",
            "risk",
            "response_text",
            "response_begin",
            "response_end",
        ]
        for key in keys:
            assert key in hallucination
        assert hallucination["hallucination_id"] == str(hallc_id)
        assert len(hallucination["response_text"]) >= 1
        hallc_id += 1



# SPDX-License-Identifier: Apache-2.0

"""
This example show how to infer or call a model using the framework
and an Ollama backend to serve the model.
"""

# Local
from granite_io import make_backend, make_io_processor
from granite_io.types import ChatCompletionInputs, UserMessage

model_name = "granite3.2:8b"
io_processor = make_io_processor(
    model_name, backend=make_backend("openai", {"model_name": model_name})
)
question = "Find the fastest way for a seller to visit all the cities in their region"
messages = [UserMessage(content=question)]
outputs = io_processor.create_chat_completion(ChatCompletionInputs(messages=messages))
print("------ WITHOUT THINKING ------")
print(outputs.results[0].next_message.content)



# SPDX-License-Identifier: Apache-2.0

"""
This example show how to infer or call a model using the framework
and a watsonx, Replicate, or Ollama backend with the LiteLLM client.
In this scenario the chat request enables thinking mode in the model
to provide better understanding of how the model arrived at its answer.
"""

# Third Party
from dotenv import load_dotenv

# Local
from granite_io import make_backend, make_io_processor
from granite_io.types import ChatCompletionInputs, UserMessage

load_dotenv()

# LiteLLM will honor environment variables.
# For this example, use a local .env file to set the variables needed
# for watsonx or Replicate or Ollama.  load_dotenv() above will read them in.
#
# Example .env entries (choose one provider):
#
# For watsonx:
#
# MODEL_NAME=watsonx/ibm/granite-3-2-8b-instruct
# WATSONX_API_BASE=https://us-south.ml.cloud.ibm.com
# WATSONX_PROJECT_ID=<your watsonx project ID>
# WATSONX_API_KEY=<your watsonx api key>
#
# For Replicate:
#
# REPLICATE_API_TOKEN=<your Replicate api key>
# MODEL_NAME=replicate/ibm-granite/granite-3.2-8b-instruct

# For Ollama:
#
# MODEL_NAME="ollama/granite3.2:8b"
# OPENAI_API_BASE=http://localhost:11434/v1
# OPENAI_API_KEY="ollama"

# NOTE! If you don't specify any of the above, the Ollama defaults are used.

model_type = "Granite 3.2"
model_name = "ollama/granite3.2:8b"
io_processor = make_io_processor(
    model_type, backend=make_backend("litellm", {"model_name": model_name})
)
question = "Find the fastest way for a seller to visit all the cities in their region"
messages = [UserMessage(content=question)]

# With Thinking
outputs = io_processor.create_chat_completion(
    ChatCompletionInputs(messages=messages, thinking=True)
)
print("------ WITH THINKING ------")
print(">> Thoughts:")
print(outputs.results[0].next_message.reasoning_content)
print(">> Response:")
print(outputs.results[0].next_message.content)
